import logging

from django.conf import settings
from django.http import HttpRequest
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import extend_schema, extend_schema_view
from rest_framework import status
from rest_framework.decorators import action

from rest_framework.permissions import IsAdminUser
from rest_framework.response import Response

from core.api.serializers import ErrorResponseSerializer
from core.api.views import BaseModelViewSet
from core.utils.serializers import (
    RequestSerializersConfig,
    ResponseSerializersConfig,
)
from core.utils.views import cache_methods
from order.filters import OrderFilter
from order.models import Order
from order.serializers.payment import (
    PaymentStatusResponseSerializer,
    ProcessPaymentRequestSerializer,
    ProcessPaymentResponseSerializer,
    RefundRequestSerializer,
    RefundResponseSerializer,
)
from pay_way.models import PayWay
from pay_way.services import PayWayService

logger = logging.getLogger(__name__)

req_serializers: RequestSerializersConfig = {
    "process_payment": ProcessPaymentRequestSerializer,
    "refund_payment": RefundRequestSerializer,
}

res_serializers: ResponseSerializersConfig = {
    "process_payment": ProcessPaymentResponseSerializer,
    "check_payment_status": PaymentStatusResponseSerializer,
    "refund_payment": RefundResponseSerializer,
}


@extend_schema_view(
    process_payment=extend_schema(
        operation_id="processOrderPayment",
        summary=_("Process payment for an order"),
        description=_(
            "Initiates payment processing for the specified order using the provided payment method."
        ),
        tags=["Order Payment"],
        request=ProcessPaymentRequestSerializer,
        responses={
            200: ProcessPaymentResponseSerializer,
            400: ErrorResponseSerializer,
            403: ErrorResponseSerializer,
            500: ErrorResponseSerializer,
        },
    ),
    check_payment_status=extend_schema(
        operation_id="checkOrderPaymentStatus",
        summary=_("Check payment status"),
        description=_(
            "Retrieves the current payment status for the specified order."
        ),
        tags=["Order Payment"],
        responses={
            200: PaymentStatusResponseSerializer,
            400: ErrorResponseSerializer,
            403: ErrorResponseSerializer,
            500: ErrorResponseSerializer,
        },
    ),
    refund_payment=extend_schema(
        operation_id="refundOrderPayment",
        summary=_("Refund payment"),
        description=_(
            "Initiates a refund for the specified order. Requires admin permissions."
        ),
        tags=["Order Payment"],
        request=RefundRequestSerializer,
        responses={
            200: RefundResponseSerializer,
            400: ErrorResponseSerializer,
            500: ErrorResponseSerializer,
        },
    ),
)
@cache_methods(settings.DEFAULT_CACHE_TTL, methods=["check_payment_status"])
class OrderPaymentViewSet(BaseModelViewSet):
    queryset = Order.objects.all()
    lookup_field = "pk"
    request_serializers = req_serializers
    response_serializers = res_serializers
    filterset_class = OrderFilter
    ordering_fields = [
        "id",
        "created_at",
        "updated_at",
        "status",
        "payment_status",
    ]
    ordering = ["-created_at"]
    search_fields = [
        "first_name",
        "last_name",
        "email",
        "payment_id",
        "tracking_number",
    ]

    def get_permissions(self):
        if self.action == "refund_payment":
            self.permission_classes = [IsAdminUser]
        return super().get_permissions()

    def check_order_permission(self, request, order):
        if request.user.is_authenticated and request.user.is_staff:
            return True

        if request.user.is_authenticated and order.user_id == request.user.id:
            return True

        if not request.user.is_authenticated:
            if not order.user_id:
                request.session[f"order_{order.uuid}"] = True
                return True

            if request.session.get(f"order_{order.uuid}"):
                return True

        return False

    def _validate_process(
        self, request: HttpRequest, order: Order
    ) -> Response | None:
        if not self.check_order_permission(request, order):
            return Response(
                {
                    "detail": _(
                        "You do not have permission to process payment for this order."
                    )
                },
                status=status.HTTP_403_FORBIDDEN,
            )

        if order.is_paid:
            return Response(
                {"detail": _("This order has already been paid.")},
                status=status.HTTP_400_BAD_REQUEST,
            )

        pay_way_id = request.data.get("pay_way_id")
        if not pay_way_id:
            return Response(
                {"detail": _("Payment method is required.")},
                status=status.HTTP_400_BAD_REQUEST,
            )

        try:
            pay_way = PayWay.objects.get(id=pay_way_id, active=True)
        except PayWay.DoesNotExist:
            return Response(
                {"detail": _("Invalid payment method.")},
                status=status.HTTP_400_BAD_REQUEST,
            )

        order.pay_way = pay_way
        order.save(update_fields=["pay_way"])
        return None

    @action(detail=True, methods=["post"])
    def process_payment(self, request: HttpRequest, pk: str) -> Response:
        order = get_object_or_404(Order, pk=pk)

        request_serializer_class = self.get_request_serializer()
        request_serializer = request_serializer_class(data=request.data)
        request_serializer.is_valid(raise_exception=True)

        validation_error = self._validate_process(request, order)
        if validation_error:
            return validation_error

        payment_data = request_serializer.validated_data.get("payment_data", {})
        try:
            success, response_data = PayWayService.process_payment(
                pay_way=order.pay_way, order=order, **payment_data
            )
            if not success:
                logger.error(
                    "Payment processing failed for order %s",
                    order.id,
                    extra={
                        "order_id": order.id,
                        "pay_way": order.pay_way.id,
                        "error": response_data.get("error", "Unknown error"),
                    },
                )
                return Response(
                    {
                        "detail": _("Payment processing failed."),
                        "error": response_data.get("error"),
                    },
                    status=status.HTTP_400_BAD_REQUEST,
                )

            response_data_formatted = {
                "detail": _("Payment processing initiated."),
                "order_id": order.id,
                "payment_status": order.payment_status,
                "payment_id": response_data.get("payment_id", ""),
                "requires_confirmation": order.pay_way.requires_confirmation,
                "is_online_payment": order.pay_way.is_online_payment,
                "provider_data": {
                    k: v
                    for k, v in response_data.items()
                    if k not in ["payment_id", "status", "error"]
                },
            }

            response_serializer_class = self.get_response_serializer()
            response_serializer = response_serializer_class(
                response_data_formatted
            )
            return Response(response_serializer.data)

        except Exception:
            logger.exception(
                "Exception during payment processing for order %s",
                order.id,
                extra={"order_id": order.id, "pay_way": order.pay_way.id},
            )
            return Response(
                {
                    "detail": _(
                        "An error occurred while processing the payment."
                    )
                },
                status=status.HTTP_500_INTERNAL_SERVER_ERROR,
            )

    @action(detail=True, methods=["get"])
    def check_payment_status(
        self, request: HttpRequest, pk: str | None = None
    ) -> Response:
        order = get_object_or_404(Order, pk=pk)

        if not self.check_order_permission(request, order):
            return Response(
                {
                    "detail": _(
                        "You do not have permission to check payment status for this order."
                    )
                },
                status=status.HTTP_403_FORBIDDEN,
            )

        if not order.pay_way:
            return Response(
                {"detail": _("This order has no payment method selected.")},
                status=status.HTTP_400_BAD_REQUEST,
            )

        if not order.payment_id:
            return Response(
                {"detail": _("This order has no associated payment.")},
                status=status.HTTP_400_BAD_REQUEST,
            )

        try:
            payment_status, status_data = PayWayService.check_payment_status(
                pay_way=order.pay_way,
                order=order,
            )

            response_data = {
                "order_id": order.id,
                "payment_status": payment_status,
                "is_paid": order.is_paid,
                "status_details": status_data,
            }

            response_serializer_class = self.get_response_serializer()
            response_serializer = response_serializer_class(response_data)
            return Response(response_serializer.data)

        except Exception:
            logger.exception(
                "Exception during payment status check for order %s",
                order.id,
                extra={"order_id": order.id},
            )
            return Response(
                {
                    "detail": _(
                        "An error occurred while checking payment status."
                    )
                },
                status=status.HTTP_500_INTERNAL_SERVER_ERROR,
            )

    @action(detail=True, methods=["post"])
    def refund_payment(
        self, request: HttpRequest, pk: str | None = None
    ) -> Response:
        order = get_object_or_404(Order, pk=pk)

        request_serializer_class = self.get_request_serializer()
        request_serializer = request_serializer_class(data=request.data)
        request_serializer.is_valid(raise_exception=True)

        if not order.is_paid:
            return Response(
                {"detail": _("This order has not been paid.")},
                status=status.HTTP_400_BAD_REQUEST,
            )

        if not order.pay_way:
            return Response(
                {"detail": _("This order has no payment method selected.")},
                status=status.HTTP_400_BAD_REQUEST,
            )

        amount = None
        validated_data = request_serializer.validated_data
        if "amount" in validated_data and "currency" in validated_data:
            from djmoney.money import Money  # noqa: PLC0415

            amount = Money(
                amount=validated_data.get("amount"),
                currency=validated_data.get("currency"),
            )

        try:
            success, response_data = PayWayService.refund_payment(
                pay_way=order.pay_way,
                order=order,
                amount=amount,
            )

            if not success:
                logger.error(
                    "Payment refund failed for order %s",
                    order.id,
                    extra={
                        "order_id": order.id,
                        "error": response_data.get("error", "Unknown error"),
                    },
                )
                return Response(
                    {
                        "detail": _("Payment refund failed."),
                        "error": response_data.get("error"),
                    },
                    status=status.HTTP_400_BAD_REQUEST,
                )

            response_data_formatted = {
                "detail": _("Payment refund initiated."),
                "order_id": order.id,
                "payment_status": order.payment_status,
                "refund_id": response_data.get("refund_id", ""),
                "refund_details": response_data,
            }

            response_serializer_class = self.get_response_serializer()
            response_serializer = response_serializer_class(
                response_data_formatted
            )
            return Response(response_serializer.data)

        except Exception:
            logger.exception(
                "Exception during payment refund for order %s",
                order.id,
                extra={"order_id": order.id},
            )
            return Response(
                {"detail": _("An error occurred while processing the refund.")},
                status=status.HTTP_500_INTERNAL_SERVER_ERROR,
            )



import os
import shutil
import tempfile
from datetime import timedelta
from pathlib import Path
from unittest.mock import Mock, mock_open, patch

import pytest
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.management import CommandError
from django.test import TestCase, override_settings
from django.utils import timezone

from cart.models import Cart
from core.tasks import (
    MonitoredTask,
    backup_database_task,
    cleanup_old_backups,
    clear_all_cache_task,
    clear_carts_for_none_users_task,
    clear_duplicate_history_task,
    clear_expired_notifications_task,
    clear_expired_sessions_task,
    clear_development_log_files_task,
    clear_old_history_task,
    monitor_system_health,
    scheduled_database_backup,
    send_inactive_user_notifications,
    track_task_metrics,
    validate_task_configuration,
)
from user.factories.account import UserAccountFactory


@pytest.mark.django_db
class TestMonitoredTask(TestCase):
    def setUp(self):
        self.task = MonitoredTask()
        self.task.name = "test_task"

    @patch("core.tasks.logger")
    def test_on_success_logs_completion(self, mock_logger):
        self.task.on_success(
            retval="success", task_id="12345", args=(), kwargs={}
        )

        mock_logger.info.assert_called_once_with(
            "Task test_task completed successfully. Task ID: 12345"
        )

    @patch("core.tasks.logger")
    def test_on_failure_logs_error(self, mock_logger):
        exception = Exception("Test error")
        self.task.on_failure(
            exc=exception, task_id="12345", args=(), kwargs={}, einfo=None
        )

        mock_logger.error.assert_called_once_with(
            "Task test_task failed. Task ID: 12345, Error: Test error"
        )


class TestTrackTaskMetrics(TestCase):
    @patch("core.tasks.logger")
    @patch("time.time")
    def test_successful_task_tracking(self, mock_time, mock_logger):
        mock_time.side_effect = [1000.0, 1001.5]

        @track_task_metrics
        def dummy_task():
            return "success"

        result = dummy_task()

        self.assertEqual(result, "success")
        mock_logger.info.assert_called_once()
        call_args = mock_logger.info.call_args
        self.assertIn("Task dummy_task completed successfully", call_args[0][0])
        self.assertEqual(call_args[1]["extra"]["duration"], 1.5)
        self.assertEqual(call_args[1]["extra"]["status"], "success")

    @patch("core.tasks.logger")
    @patch("time.time")
    def test_failed_task_tracking(self, mock_time, mock_logger):
        mock_time.side_effect = [1000.0, 1001.5]

        @track_task_metrics
        def failing_task():
            raise ValueError("Test error")

        with self.assertRaises(ValueError):
            failing_task()

        mock_logger.error.assert_called_once()
        call_args = mock_logger.error.call_args
        self.assertIn("Task failing_task failed", call_args[0][0])
        self.assertEqual(call_args[1]["extra"]["duration"], 1.5)
        self.assertEqual(call_args[1]["extra"]["status"], "failure")
        self.assertEqual(call_args[1]["extra"]["error"], "Test error")


@pytest.mark.django_db
class TestClearExpiredSessionsTask(TestCase):
    @patch("core.tasks.management.call_command")
    @patch("core.tasks.logger")
    def test_successful_session_cleanup(self, mock_logger, mock_call_command):
        result = clear_expired_sessions_task()

        mock_call_command.assert_called_once_with("clearsessions", verbosity=0)
        self.assertEqual(result["status"], "success")
        self.assertEqual(result["message"], "All expired sessions deleted")
        mock_logger.info.assert_any_call("Starting expired sessions cleanup")

    @patch("core.tasks.management.call_command")
    @patch("core.tasks.logger")
    def test_command_error_handling(self, mock_logger, mock_call_command):
        mock_call_command.side_effect = CommandError("Command failed")

        result = clear_expired_sessions_task()

        self.assertEqual(result["status"], "error")
        self.assertEqual(result["error_type"], "CommandError")
        self.assertIn("Command failed", result["message"])
        mock_logger.error.assert_called()

    @patch("core.tasks.management.call_command")
    @patch("core.tasks.logger")
    def test_unexpected_error_handling(self, mock_logger, mock_call_command):
        mock_call_command.side_effect = RuntimeError("Unexpected error")

        with self.assertRaises(RuntimeError):
            clear_expired_sessions_task()

        mock_logger.exception.assert_called_with(
            "Unexpected error in clear_expired_sessions"
        )


@pytest.mark.django_db
class TestClearAllCacheTask(TestCase):
    @patch("core.tasks.management.call_command")
    @patch("core.tasks.logger")
    def test_successful_cache_cleanup(self, mock_logger, mock_call_command):
        result = clear_all_cache_task()

        mock_call_command.assert_called_once_with("clear_cache", verbosity=0)
        self.assertEqual(result["status"], "success")
        self.assertEqual(result["message"], "All cache deleted")

    @patch("core.tasks.management.call_command")
    def test_cache_cleanup_command_error(self, mock_call_command):
        mock_call_command.side_effect = CommandError("Cache error")

        result = clear_all_cache_task()

        self.assertEqual(result["status"], "error")
        self.assertEqual(result["error_type"], "CommandError")


@pytest.mark.django_db
class TestClearDuplicateHistoryTask(TestCase):
    @patch("core.tasks.management.call_command")
    @patch("core.tasks.logger")
    def test_duplicate_history_cleanup_with_defaults(
        self, mock_logger, mock_call_command
    ):
        result = clear_duplicate_history_task()

        mock_call_command.assert_called_once_with(
            "clean_duplicate_history", "--auto"
        )
        self.assertEqual(result["status"], "success")
        self.assertIn("Duplicate history entries cleaned", result["message"])

    @patch("core.tasks.management.call_command")
    def test_duplicate_history_cleanup_with_parameters(self, mock_call_command):
        excluded_fields = ["field1", "field2"]
        minutes = 30

        result = clear_duplicate_history_task(
            excluded_fields=excluded_fields, minutes=minutes
        )

        mock_call_command.assert_called_once_with(
            "clean_duplicate_history",
            "-m",
            "30",
            "--excluded_fields",
            "field1",
            "field2",
            "--auto",
        )
        self.assertEqual(result["status"], "success")
        self.assertEqual(
            result["parameters"]["excluded_fields"], excluded_fields
        )
        self.assertEqual(result["parameters"]["minutes"], minutes)

    @patch("core.tasks.management.call_command")
    def test_duplicate_history_cleanup_command_error(self, mock_call_command):
        mock_call_command.side_effect = CommandError("History error")

        result = clear_duplicate_history_task()

        self.assertEqual(result["status"], "error")
        self.assertEqual(result["error_type"], "CommandError")


@pytest.mark.django_db
class TestClearOldHistoryTask(TestCase):
    @patch("core.tasks.management.call_command")
    @patch("core.tasks.logger")
    def test_old_history_cleanup_with_default_days(
        self, mock_logger, mock_call_command
    ):
        result = clear_old_history_task()

        mock_call_command.assert_called_once_with(
            "clean_old_history", "--days", "365", "--auto"
        )
        self.assertEqual(result["status"], "success")
        self.assertIn(
            "Old history entries cleaned (older than 365 days)",
            result["message"],
        )

    @patch("core.tasks.management.call_command")
    def test_old_history_cleanup_with_custom_days(self, mock_call_command):
        result = clear_old_history_task(days=30)

        mock_call_command.assert_called_once_with(
            "clean_old_history", "--days", "30", "--auto"
        )
        self.assertEqual(result["status"], "success")


@pytest.mark.django_db
class TestClearExpiredNotificationsTask(TestCase):
    @patch("core.tasks.management.call_command")
    @patch("core.tasks.logger")
    def test_expired_notifications_cleanup(
        self, mock_logger, mock_call_command
    ):
        result = clear_expired_notifications_task(days=30)

        mock_call_command.assert_called_once_with(
            "expire_notifications", "--days", "30"
        )
        self.assertEqual(result["status"], "success")
        self.assertEqual(result["message"], "Expired notifications deleted")

    @patch("core.tasks.management.call_command")
    def test_expired_notifications_with_default_days(self, mock_call_command):
        result = clear_expired_notifications_task()

        mock_call_command.assert_called_once_with(
            "expire_notifications", "--days", "365"
        )
        self.assertEqual(result["status"], "success")


@pytest.mark.django_db
class TestClearCartsForNoneUsersTask(TestCase):
    def setUp(self):
        self.user = UserAccountFactory()
        self.user_cart = Cart.objects.create(
            user=self.user, session_key="user_session"
        )
        self.null_cart1 = Cart.objects.create(
            user=None, session_key="null_session1"
        )
        self.null_cart2 = Cart.objects.create(
            user=None, session_key="null_session2"
        )

    @patch("core.tasks.logger")
    def test_clear_null_user_carts(self, mock_logger):
        result = clear_carts_for_none_users_task()

        self.assertEqual(result["status"], "success")
        self.assertEqual(result["carts_deleted"], 2)

        self.assertFalse(Cart.objects.filter(user=None).exists())
        self.assertTrue(Cart.objects.filter(user=self.user).exists())

        mock_logger.info.assert_called()

    @patch("core.tasks.logger")
    def test_clear_null_user_carts_no_carts(self, mock_logger):
        Cart.objects.filter(user=None).delete()

        result = clear_carts_for_none_users_task()

        self.assertEqual(result["status"], "success")
        self.assertEqual(result["carts_deleted"], 0)

    @patch("core.tasks.transaction.atomic")
    @patch("core.tasks.logger")
    def test_clear_null_user_carts_database_error(
        self, mock_logger, mock_atomic
    ):
        mock_atomic.side_effect = Exception("Database error")

        with self.assertRaises(Exception):
            clear_carts_for_none_users_task()

        mock_logger.exception.assert_called_with(
            "Error clearing null user carts"
        )


@pytest.mark.django_db
class TestClearLogFilesTask(TestCase):
    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()
        self.logs_path = os.path.join(self.temp_dir, "logs")
        os.makedirs(self.logs_path)

    def tearDown(self):
        shutil.rmtree(self.temp_dir, ignore_errors=True)

    @override_settings(BASE_DIR=Path(__file__).parent)
    @patch("core.tasks.os.getenv")
    @patch("core.tasks.os.path.exists")
    @patch("core.tasks.os.path.join")
    @patch("core.tasks.os.listdir")
    @patch("core.tasks.os.path.isfile")
    @patch("core.tasks.os.path.getmtime")
    @patch("core.tasks.os.path.getsize")
    @patch("core.tasks.os.remove")
    @patch("core.tasks.logger")
    def test_clear_old_log_files(
        self,
        mock_logger,
        mock_remove,
        mock_getsize,
        mock_getmtime,
        mock_isfile,
        mock_listdir,
        mock_join,
        mock_exists,
        mock_getenv,
    ):
        mock_getenv.side_effect = lambda key, default=None: {
            "KUBERNETES_SERVICE_HOST": None,
        }.get(key, default)

        def mock_exists_side_effect(path):
            if path == "/.dockerenv":
                return True
            elif path == self.logs_path:
                return True
            return False

        mock_exists.side_effect = mock_exists_side_effect
        mock_join.return_value = self.logs_path
        mock_listdir.return_value = [
            "old_log.log.1",
            "recent_log.log.2",
            "current_log.log",
            "not_a_file",
        ]

        old_time = (timezone.now() - timedelta(days=40)).timestamp()
        recent_time = (timezone.now() - timedelta(days=10)).timestamp()

        def mock_isfile_side_effect(path):
            return "not_a_file" not in path

        def mock_getmtime_side_effect(path):
            if "old_log.log.1" in path:
                return old_time
            return recent_time

        def mock_getsize_side_effect(path):
            return 1024 if "old_log.log.1" in path else 512

        mock_isfile.side_effect = mock_isfile_side_effect
        mock_getmtime.side_effect = mock_getmtime_side_effect
        mock_getsize.side_effect = mock_getsize_side_effect

        result = clear_development_log_files_task(days=30)

        self.assertEqual(result["status"], "success")
        self.assertIn("deleted_count", result)
        mock_getmtime.assert_called()

    @override_settings(BASE_DIR=Path(__file__).parent)
    @patch("core.tasks.os.getenv")
    @patch("core.tasks.os.path.exists")
    @patch("core.tasks.os.path.join")
    @patch("core.tasks.logger")
    def test_clear_log_files_no_logs_directory(
        self, mock_logger, mock_join, mock_exists, mock_getenv
    ):
        mock_getenv.side_effect = lambda key, default=None: {
            "KUBERNETES_SERVICE_HOST": None,
        }.get(key, default)

        def mock_exists_side_effect(path):
            if path == "/.dockerenv":
                return True
            elif path == "/nonexistent/logs":
                return False
            return False

        mock_exists.side_effect = mock_exists_side_effect
        mock_join.return_value = "/nonexistent/logs"

        result = clear_development_log_files_task()

        self.assertEqual(result["status"], "skipped")
        self.assertEqual(result["reason"], "logs directory not found")
        mock_logger.warning.assert_called()

    @patch("core.tasks.os.getenv")
    @patch("core.tasks.os.path.exists")
    def test_clear_log_files_kubernetes_environment(
        self, mock_exists, mock_getenv
    ):
        mock_getenv.side_effect = lambda key, default=None: {
            "KUBERNETES_SERVICE_HOST": "10.0.0.1",
        }.get(key, default)

        result = clear_development_log_files_task()

        self.assertEqual(result["status"], "skipped")
        self.assertEqual(result["reason"], "kubernetes environment detected")

    @patch("core.tasks.os.getenv")
    @patch("core.tasks.os.path.exists")
    def test_clear_log_files_non_docker_environment(
        self, mock_exists, mock_getenv
    ):
        mock_getenv.side_effect = lambda key, default=None: {
            "KUBERNETES_SERVICE_HOST": None,
        }.get(key, default)

        def mock_exists_side_effect(path):
            if path == "/.dockerenv":
                return False
            return False

        mock_exists.side_effect = mock_exists_side_effect

        result = clear_development_log_files_task()

        self.assertEqual(result["status"], "skipped")
        self.assertEqual(result["reason"], "not in docker environment")

    @override_settings(BASE_DIR=Path(__file__).parent)
    @patch("core.tasks.os.getenv")
    @patch("core.tasks.os.path.exists")
    @patch("core.tasks.os.path.join")
    @patch("core.tasks.os.listdir")
    @patch("core.tasks.logger")
    def test_clear_log_files_os_error(
        self, mock_logger, mock_listdir, mock_join, mock_exists, mock_getenv
    ):
        mock_getenv.side_effect = lambda key, default=None: {
            "KUBERNETES_SERVICE_HOST": None,
        }.get(key, default)

        def mock_exists_side_effect(path):
            if path == "/.dockerenv":
                return True
            elif path == self.logs_path:
                return True
            return False

        mock_exists.side_effect = mock_exists_side_effect
        mock_join.return_value = self.logs_path
        mock_listdir.side_effect = OSError("Permission denied")

        with self.assertRaises(OSError):
            clear_development_log_files_task()

        mock_logger.exception.assert_called()


@pytest.mark.django_db
class TestSendInactiveUserNotificationsTask(TestCase):
    def setUp(self):
        self.active_user = UserAccountFactory(
            last_login=timezone.now() - timedelta(days=30),
            is_active=True,
            email="active@example.com",
        )
        self.inactive_user = UserAccountFactory(
            last_login=timezone.now() - timedelta(days=70),
            is_active=True,
            email="inactive@example.com",
            first_name="John",
        )
        self.very_old_user = UserAccountFactory(
            last_login=timezone.now() - timedelta(days=100),
            is_active=True,
            email="old@example.com",
        )
        self.inactive_no_email = UserAccountFactory(
            last_login=timezone.now() - timedelta(days=70),
            is_active=True,
            email="",
        )

    @patch("core.tasks.send_mail")
    @patch("core.tasks.render_to_string")
    @patch("core.tasks.logger")
    @override_settings(
        NUXT_BASE_URL="https://example.com",
        DEFAULT_FROM_EMAIL="noreply@example.com",
    )
    def test_send_inactive_user_notifications_success(
        self, mock_logger, mock_render, mock_send_mail
    ):
        mock_render.return_value = "<html>Test email</html>"

        result = send_inactive_user_notifications()

        self.assertEqual(result["status"], "completed")
        self.assertEqual(result["total_users"], 2)
        self.assertEqual(result["emails_sent"], 2)
        self.assertEqual(result["failed"], 0)

        self.assertEqual(mock_send_mail.call_count, 2)

        call_args = mock_send_mail.call_args_list[0][1]
        self.assertIn("We miss you!", call_args["subject"])
        self.assertEqual(call_args["from_email"], "noreply@example.com")

    @patch("core.tasks.send_mail")
    @patch("core.tasks.render_to_string")
    @patch("core.tasks.logger")
    def test_send_inactive_user_notifications_with_failures(
        self, mock_logger, mock_render, mock_send_mail
    ):
        mock_render.return_value = "<html>Test email</html>"
        mock_send_mail.side_effect = [
            Exception("SMTP error"),
            None,
        ]

        result = send_inactive_user_notifications()

        self.assertEqual(result["status"], "completed")
        self.assertEqual(result["total_users"], 2)
        self.assertEqual(result["emails_sent"], 1)
        self.assertEqual(result["failed"], 1)
        self.assertEqual(len(result["failed_details"]), 1)

    @patch("core.tasks.send_mail")
    @patch("core.tasks.render_to_string")
    @patch("core.tasks.logger")
    def test_send_inactive_user_notifications_too_many_failures(
        self, mock_logger, mock_render, mock_send_mail
    ):
        mock_render.return_value = "<html>Test email</html>"
        mock_send_mail.side_effect = Exception("SMTP error")

        for i in range(55):
            UserAccountFactory(
                last_login=timezone.now() - timedelta(days=70),
                is_active=True,
                email=f"inactive{i}@example.com",
            )

        result = send_inactive_user_notifications()

        self.assertEqual(result["emails_sent"], 0)
        self.assertEqual(len(result["failed_details"]), 10)
        mock_logger.error.assert_any_call(
            "Too many email failures, stopping task"
        )


@pytest.mark.django_db
class TestMonitorSystemHealthTask(TestCase):
    @patch("core.tasks.connections")
    @patch("core.tasks.cache")
    @patch("core.tasks.open", new_callable=mock_open)
    @patch("core.tasks.os.remove")
    @patch("core.tasks.logger")
    @override_settings(MEDIA_ROOT="/tmp/media")
    def test_monitor_system_health_all_passed(
        self, mock_logger, mock_remove, mock_file, mock_cache, mock_connections
    ):
        mock_cursor = Mock()
        mock_connections.__getitem__.return_value.cursor.return_value.__enter__.return_value = mock_cursor

        mock_cache.set.return_value = None
        mock_cache.get.return_value = "ok"

        result = monitor_system_health()

        self.assertEqual(result["status"], "healthy")
        self.assertTrue(result["checks"]["database"])
        self.assertTrue(result["checks"]["cache"])
        self.assertTrue(result["checks"]["storage"])
        self.assertEqual(len(result["errors"]), 0)

    @patch("core.tasks.connections")
    @patch("core.tasks.cache")
    @patch("core.tasks.open")
    @patch("core.tasks.send_mail")
    @patch("core.tasks.logger")
    @override_settings(
        MEDIA_ROOT="/tmp/media",
        DEFAULT_FROM_EMAIL="admin@example.com",
        ADMIN_EMAIL="admin@example.com",
    )
    def test_monitor_system_health_database_failure(
        self,
        mock_logger,
        mock_send_mail,
        mock_file,
        mock_cache,
        mock_connections,
    ):
        mock_connections.__getitem__.return_value.cursor.side_effect = (
            Exception("DB error")
        )

        mock_cache.set.return_value = None
        mock_cache.get.return_value = "ok"

        with self.assertRaises(
            Exception,
            msg="Critical system health check failed",
        ):
            monitor_system_health()

        mock_send_mail.assert_called_once()
        call_args = mock_send_mail.call_args[1]
        self.assertIn(
            "CRITICAL: System Health Check Failed", call_args["subject"]
        )

    @patch("core.tasks.connections")
    @patch("core.tasks.cache")
    @patch("core.tasks.open", new_callable=mock_open)
    @patch("core.tasks.os.remove")
    @patch("core.tasks.logger")
    @override_settings(MEDIA_ROOT="/tmp/media")
    def test_monitor_system_health_cache_failure(
        self, mock_logger, mock_remove, mock_file, mock_cache, mock_connections
    ):
        mock_cursor = Mock()
        mock_connections.__getitem__.return_value.cursor.return_value.__enter__.return_value = mock_cursor

        mock_cache.set.side_effect = Exception("Cache error")

        result = monitor_system_health()

        self.assertEqual(result["status"], "degraded")
        self.assertTrue(result["checks"]["database"])
        self.assertFalse(result["checks"]["cache"])
        self.assertTrue(result["checks"]["storage"])
        self.assertGreater(len(result["errors"]), 0)

    @patch("core.tasks.connections")
    @patch("core.tasks.cache")
    @patch("core.tasks.logger")
    @override_settings(MEDIA_ROOT="/tmp/media")
    def test_monitor_system_health_storage_failure(
        self, mock_logger, mock_cache, mock_connections
    ):
        mock_cursor = Mock()
        mock_connections.__getitem__.return_value.cursor.return_value.__enter__.return_value = mock_cursor

        mock_cache.set.return_value = None
        mock_cache.get.return_value = "ok"

        mock_open_obj = mock_open()
        mock_open_obj.side_effect = OSError("Storage error")

        with patch("core.tasks.open", mock_open_obj):
            result = monitor_system_health()

        self.assertEqual(result["status"], "degraded")
        self.assertTrue(result["checks"]["database"])
        self.assertTrue(result["checks"]["cache"])
        self.assertFalse(result["checks"]["storage"])


@pytest.mark.django_db
class TestBackupDatabaseTask(TestCase):
    @patch("core.tasks.management.call_command")
    @patch("core.tasks.Path")
    @patch("core.tasks.logger")
    @override_settings(BASE_DIR="/test/base")
    def test_backup_database_success(
        self, mock_logger, mock_path, mock_call_command
    ):
        mock_backup_dir = Mock()
        mock_backup_file = Mock()
        mock_backup_file.name = "backup_20231201_120000.sql"
        mock_backup_file.stat.return_value.st_size = 1024000
        mock_backup_file.stat.return_value.st_ctime = timezone.now().timestamp()

        mock_backup_dir.exists.return_value = True
        mock_backup_dir.iterdir.return_value = [mock_backup_file]
        mock_path.return_value.__truediv__.return_value = mock_backup_dir

        result = backup_database_task(
            output_dir="backups",
            filename="test_backup",
            compress=True,
            format_type="custom",
        )

        self.assertEqual(result["status"], "success")
        self.assertEqual(result["file_size"], 1024000)
        self.assertIn(
            "Database backup completed successfully", result["result_message"]
        )

        mock_call_command.assert_called_once_with(
            "backup_database",
            "--output-dir",
            "backups",
            "--filename",
            "test_backup",
            "--compress",
            "--format",
            "custom",
        )

    @patch("core.tasks.management.call_command")
    @patch("core.tasks.logger")
    def test_backup_database_command_error(
        self, mock_logger, mock_call_command
    ):
        mock_call_command.side_effect = CommandError("Backup failed")

        result = backup_database_task()

        self.assertEqual(result["status"], "error")
        self.assertEqual(result["error_type"], "CommandError")
        self.assertIn("Backup failed", result["result_message"])

    @patch("core.tasks.management.call_command")
    @patch("core.tasks.Path")
    @patch("core.tasks.logger")
    @override_settings(BASE_DIR="/test/base")
    def test_backup_database_no_backup_dir(
        self, mock_logger, mock_path, mock_call_command
    ):
        mock_backup_dir = Mock()
        mock_backup_dir.exists.return_value = False
        mock_path.return_value.__truediv__.return_value = mock_backup_dir

        result = backup_database_task()

        self.assertEqual(result["status"], "success")
        self.assertEqual(result["file_size"], 0)
        self.assertIsNone(result["backup_file"])


@pytest.mark.django_db
class TestScheduledDatabaseBackupTask(TestCase):
    @patch("core.tasks.backup_database_task")
    @patch("core.tasks.cleanup_old_backups")
    @patch("core.tasks.logger")
    def test_scheduled_backup_success(
        self, mock_logger, mock_cleanup, mock_backup
    ):
        mock_result = Mock()
        mock_result.result = {
            "status": "success",
            "message": "Backup completed",
        }
        mock_backup.apply.return_value = mock_result

        result = scheduled_database_backup()

        self.assertEqual(result["status"], "success")
        mock_backup.apply.assert_called_once()
        mock_cleanup.delay.assert_called_once_with(
            days=7, backup_dir="backups/scheduled"
        )

    @patch("core.tasks.backup_database_task")
    @patch("core.tasks.send_mail")
    @patch("core.tasks.logger")
    @override_settings(
        DEFAULT_FROM_EMAIL="admin@example.com", ADMIN_EMAIL="admin@example.com"
    )
    def test_scheduled_backup_failure(
        self, mock_logger, mock_send_mail, mock_backup
    ):
        mock_result = Mock()
        mock_result.result = {"status": "error", "message": "Backup failed"}
        mock_backup.apply.return_value = mock_result

        with self.assertRaises(Exception):
            scheduled_database_backup()

        mock_send_mail.assert_called_once()
        call_args = mock_send_mail.call_args[1]
        self.assertIn(
            "ALERT: Scheduled Database Backup Failed", call_args["subject"]
        )

    @patch("core.tasks.backup_database_task")
    @patch("core.tasks.send_mail")
    @patch("core.tasks.logger")
    def test_scheduled_backup_email_failure(
        self, mock_logger, mock_send_mail, mock_backup
    ):
        mock_result = Mock()
        mock_result.result = {"status": "error", "message": "Backup failed"}
        mock_backup.apply.return_value = mock_result

        mock_send_mail.side_effect = Exception("Email error")

        with self.assertRaises(Exception):
            scheduled_database_backup()

        mock_logger.error.assert_any_call(
            "Failed to send backup failure alert email: Email error"
        )


@pytest.mark.django_db
class TestCleanupOldBackupsTask(TestCase):
    @patch("core.tasks.Path")
    @patch("core.tasks.logger")
    @override_settings(BASE_DIR="/test/base")
    def test_cleanup_old_backups_success(self, mock_logger, mock_path):
        mock_backup_dir = Mock()
        mock_backup_dir.exists.return_value = True

        old_file = Mock()
        old_file.is_file.return_value = True
        old_file.name = "old_backup.sql"
        old_file.stat.return_value.st_mtime = (
            timezone.now() - timedelta(days=40)
        ).timestamp()
        old_file.stat.return_value.st_size = 1024

        recent_file = Mock()
        recent_file.is_file.return_value = True
        recent_file.name = "recent_backup.sql"
        recent_file.stat.return_value.st_mtime = (
            timezone.now() - timedelta(days=10)
        ).timestamp()

        mock_backup_dir.iterdir.return_value = [old_file, recent_file]
        mock_path.return_value.__truediv__.return_value = mock_backup_dir

        result = cleanup_old_backups(days=30, backup_dir="backups")

        self.assertEqual(result["status"], "success")
        self.assertEqual(result["deleted_count"], 1)
        self.assertEqual(result["total_size_freed"], 1024)
        old_file.unlink.assert_called_once()
        recent_file.unlink.assert_not_called()

    @patch("core.tasks.Path")
    @patch("core.tasks.logger")
    @override_settings(BASE_DIR="/test/base")
    def test_cleanup_old_backups_no_directory(self, mock_logger, mock_path):
        mock_backup_dir = Mock()
        mock_backup_dir.exists.return_value = False
        mock_path.return_value.__truediv__.return_value = mock_backup_dir

        result = cleanup_old_backups()

        self.assertEqual(result["status"], "skipped")
        self.assertEqual(result["reason"], "backup directory not found")
        mock_logger.warning.assert_called()

    @patch("core.tasks.Path")
    @patch("core.tasks.logger")
    @override_settings(BASE_DIR="/test/base")
    def test_cleanup_old_backups_file_error(self, mock_logger, mock_path):
        mock_backup_dir = Mock()
        mock_backup_dir.exists.return_value = True

        error_file = Mock()
        error_file.is_file.return_value = True
        error_file.name = "error_backup.sql"
        error_file.stat.side_effect = OSError("Permission denied")

        mock_backup_dir.iterdir.return_value = [error_file]
        mock_path.return_value.__truediv__.return_value = mock_backup_dir

        result = cleanup_old_backups()

        self.assertEqual(result["status"], "success")
        self.assertEqual(result["deleted_count"], 0)
        mock_logger.error.assert_called()


class TestValidateTaskConfiguration(TestCase):
    @override_settings(
        DEFAULT_FROM_EMAIL="test@example.com",
        ADMIN_EMAIL="admin@example.com",
        MEDIA_ROOT="/tmp/media",
        BASE_DIR="/tmp/base",
    )
    @patch("core.tasks.logger")
    def test_validate_task_configuration_success(self, mock_logger):
        validate_task_configuration()

        mock_logger.info.assert_called_with(
            "Task configuration validated successfully"
        )

    @override_settings()
    def test_validate_task_configuration_missing_settings(self):
        if hasattr(settings, "DEFAULT_FROM_EMAIL"):
            del settings.DEFAULT_FROM_EMAIL
        if hasattr(settings, "ADMIN_EMAIL"):
            del settings.ADMIN_EMAIL

        with self.assertRaises(ImproperlyConfigured) as cm:
            validate_task_configuration()

        self.assertIn("Missing required settings", str(cm.exception))


@pytest.mark.django_db
class TestTaskIntegration(TestCase):
    def setUp(self):
        self.user = UserAccountFactory()
        self.cart = Cart.objects.create(user=None, session_key="test_session")

    @patch("core.tasks.management.call_command")
    def test_multiple_cleanup_tasks_workflow(self, mock_call_command):
        session_result = clear_expired_sessions_task()
        cache_result = clear_all_cache_task()
        cart_result = clear_carts_for_none_users_task()

        self.assertEqual(session_result["status"], "success")
        self.assertEqual(cache_result["status"], "success")
        self.assertEqual(cart_result["status"], "success")

        self.assertFalse(Cart.objects.filter(user=None).exists())

        expected_calls = ["clearsessions", "clear_cache"]
        actual_calls = [call[0][0] for call in mock_call_command.call_args_list]
        for expected_call in expected_calls:
            self.assertIn(expected_call, actual_calls)



# Generated by Django 5.1 on 2024-08-24 22:51
import django.core.validators
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("country", "0004_alter_country_options"),
    ]

    operations = [
        migrations.AlterField(
            model_name="country",
            name="alpha_2",
            field=models.CharField(
                max_length=2,
                primary_key=True,
                serialize=False,
                unique=True,
                validators=[
                    django.core.validators.RegexValidator(
                        message="Enter a valid 2-letter country code.",
                        regex="^[A-Z]{2}$",
                    )
                ],
                verbose_name="Country Code Alpha 2",
            ),
        ),
        migrations.AlterField(
            model_name="country",
            name="alpha_3",
            field=models.CharField(
                max_length=3,
                unique=True,
                validators=[
                    django.core.validators.RegexValidator(
                        message="Enter a valid 3-letter country code.",
                        regex="^[A-Z]{3}$",
                    )
                ],
                verbose_name="Country Code Alpha 3",
            ),
        ),
        migrations.AlterField(
            model_name="country",
            name="sort_order",
            field=models.IntegerField(
                editable=False, null=True, verbose_name="Sort Order"
            ),
        ),
        migrations.AddIndex(
            model_name="country",
            index=models.Index(
                fields=["alpha_2"], name="country_cou_alpha_2_398383_idx"
            ),
        ),
        migrations.AddIndex(
            model_name="country",
            index=models.Index(
                fields=["alpha_3"], name="country_cou_alpha_3_c813cd_idx"
            ),
        ),
    ]



# Generated by Django 5.0.4 on 2024-04-11 11:43
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("blog", "0007_remove_blogcomment_unique_blog_comment"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="blogcategory",
            options={
                "ordering": ["-sort_order"],
                "verbose_name": "Blog Category",
                "verbose_name_plural": "Blog Categories",
            },
        ),
        migrations.AlterModelOptions(
            name="blogtag",
            options={
                "ordering": ["-sort_order"],
                "verbose_name": "Blog Tag",
                "verbose_name_plural": "Blog Tags",
            },
        ),
    ]



import importlib

import factory
from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model

from blog.models.author import BlogAuthor

available_languages = [
    lang["code"] for lang in settings.PARLER_LANGUAGES[settings.SITE_ID]
]

User = get_user_model()


def get_or_create_user():
    if User.objects.exists():
        user = User.objects.order_by("?").first()
    else:
        user_factory_module = importlib.import_module("user.factories.account")
        user_factory_class = user_factory_module.UserAccountFactory
        user = user_factory_class.create()
    return user


class BlogAuthorTranslationFactory(factory.django.DjangoModelFactory):
    language_code = factory.Iterator(available_languages)
    bio = factory.Faker("paragraph")
    master = factory.SubFactory("blog.factories.author.BlogAuthorFactory")

    class Meta:
        model = apps.get_model("blog", "BlogAuthorTranslation")
        django_get_or_create = ("language_code", "master")


class BlogAuthorFactory(factory.django.DjangoModelFactory):
    user = factory.LazyFunction(get_or_create_user)
    website = factory.Faker("url")

    class Meta:
        model = BlogAuthor
        django_get_or_create = ("user",)
        skip_postgeneration_save = True

    @factory.post_generation
    def translations(self, create, extracted, **kwargs):
        if not create:
            return

        translations = extracted or [
            BlogAuthorTranslationFactory(language_code=lang, master=self)
            for lang in available_languages
        ]

        for translation in translations:
            translation.master = self
            translation.save()



from django.contrib.auth import get_user_model
from django.db import transaction
from django.db.models.signals import post_save
from django.test import TestCase, TransactionTestCase

from user.models.subscription import SubscriptionTopic, UserSubscription
from user.signals import create_default_subscriptions

User = get_user_model()


class CreateDefaultSubscriptionsSignalTest(TransactionTestCase):
    def setUp(self):
        self.default_topics = []
        self.non_default_topics = []

        for i in range(3):
            topic = SubscriptionTopic.objects.create(
                slug=f"default-topic-{i}",
                name=f"Default Topic {i}",
                category=SubscriptionTopic.TopicCategory.NEWSLETTER,
                is_active=True,
                is_default=True,
            )
            self.default_topics.append(topic)

        for i in range(2):
            topic = SubscriptionTopic.objects.create(
                slug=f"non-default-topic-{i}",
                name=f"Non-Default Topic {i}",
                category=SubscriptionTopic.TopicCategory.PRODUCT,
                is_active=True,
                is_default=False,
            )
            self.non_default_topics.append(topic)

        self.inactive_default = SubscriptionTopic.objects.create(
            slug="inactive-default",
            name="Inactive Default Topic",
            category=SubscriptionTopic.TopicCategory.SYSTEM,
            is_active=False,
            is_default=True,
        )

    def test_new_user_gets_default_subscriptions(self):
        user = User.objects.create_user(
            email="newuser@test.com", password="testpass123", username="newuser"
        )

        subscriptions = UserSubscription.objects.filter(user=user)
        self.assertEqual(subscriptions.count(), 3)

        subscribed_topics = set(sub.topic for sub in subscriptions)
        expected_topics = set(self.default_topics)
        self.assertEqual(subscribed_topics, expected_topics)

        for subscription in subscriptions:
            self.assertEqual(
                subscription.status, UserSubscription.SubscriptionStatus.ACTIVE
            )

    def test_existing_user_no_subscriptions(self):
        post_save.disconnect(create_default_subscriptions, sender=User)

        user = User.objects.create_user(
            email="existing@test.com",
            password="testpass123",
            username="existing",
        )

        post_save.connect(create_default_subscriptions, sender=User)

        self.assertEqual(UserSubscription.objects.filter(user=user).count(), 0)

        user.first_name = "Updated"
        user.save()

        self.assertEqual(UserSubscription.objects.filter(user=user).count(), 0)

    def test_signal_only_on_create(self):
        user = User.objects.create_user(
            email="testcreate@test.com",
            password="testpass123",
            username="testcreate",
        )

        initial_count = UserSubscription.objects.filter(user=user).count()
        self.assertEqual(initial_count, 3)

        new_default = SubscriptionTopic.objects.create(
            slug="new-default",
            name="New Default Topic",
            is_active=True,
            is_default=True,
        )

        user.first_name = "Updated"
        user.save()

        final_count = UserSubscription.objects.filter(user=user).count()
        self.assertEqual(final_count, initial_count)

        self.assertFalse(
            UserSubscription.objects.filter(
                user=user, topic=new_default
            ).exists()
        )

    def test_inactive_default_topics_ignored(self):
        user = User.objects.create_user(
            email="inactivetest@test.com",
            password="testpass123",
            username="inactivetest",
        )

        self.assertFalse(
            UserSubscription.objects.filter(
                user=user, topic=self.inactive_default
            ).exists()
        )

    def test_signal_handles_errors_gracefully(self):
        SubscriptionTopic.objects.create(
            slug="problematic",
            name="Problematic Topic",
            is_active=True,
            is_default=True,
        )

        User.objects.create_user(
            email="errortest@test.com",
            password="testpass123",
            username="errortest",
        )

        self.assertTrue(
            User.objects.filter(email="errortest@test.com").exists()
        )

    def test_bulk_create_users(self):
        users_data = [
            User(email=f"bulk{i}@test.com", username=f"bulk{i}")
            for i in range(3)
        ]

        created_users = User.objects.bulk_create(users_data)

        for user in created_users:
            self.assertEqual(
                UserSubscription.objects.filter(user=user).count(), 0
            )

    def test_signal_with_transaction(self):
        with transaction.atomic():
            user = User.objects.create_user(
                email="transactiontest@test.com",
                password="testpass123",
                username="transactiontest",
            )

            subscriptions = UserSubscription.objects.filter(user=user)
            self.assertEqual(subscriptions.count(), 3)

        self.assertEqual(UserSubscription.objects.filter(user=user).count(), 3)


class SignalDisconnectionTest(TestCase):
    def setUp(self):
        self.default_topic = SubscriptionTopic.objects.create(
            slug="default-topic",
            name="Default Topic",
            is_active=True,
            is_default=True,
        )

    def test_user_creation_without_signal(self):
        post_save.disconnect(create_default_subscriptions, sender=User)

        try:
            user = User.objects.create_user(
                email="nosignal@test.com",
                password="testpass123",
                username="nosignal",
            )

            self.assertEqual(
                UserSubscription.objects.filter(user=user).count(), 0
            )

        finally:
            post_save.connect(create_default_subscriptions, sender=User)

    def test_manual_subscription_creation(self):
        post_save.disconnect(create_default_subscriptions, sender=User)

        try:
            user = User.objects.create_user(
                email="manual@test.com",
                password="testpass123",
                username="manual",
            )

            default_topics = SubscriptionTopic.objects.filter(
                is_active=True, is_default=True
            )

            for topic in default_topics:
                UserSubscription.objects.create(
                    user=user,
                    topic=topic,
                    status=UserSubscription.SubscriptionStatus.ACTIVE,
                )

            self.assertEqual(
                UserSubscription.objects.filter(user=user).count(),
                default_topics.count(),
            )

        finally:
            post_save.connect(create_default_subscriptions, sender=User)



from unittest import TestCase
from unittest.mock import MagicMock, Mock, patch

from order.enum.status import OrderStatus
from order.signals.handlers import handle_order_post_save


class OrderSignalHandlersTestCase(TestCase):
    def setUp(self):
        self.order = Mock()
        self.order.id = 1
        self.order.status = OrderStatus.PENDING
        self.order._previous_status = None

        self.sender = MagicMock()

    @patch("order.signals.order_status_changed.send")
    @patch("order.signals.order_created.send")
    def test_handle_order_post_save_new_order(
        self, mock_order_created, mock_signal
    ):
        handle_order_post_save(
            sender=self.sender, instance=self.order, created=True
        )

        mock_signal.assert_not_called()

        mock_order_created.assert_called_once_with(
            sender=self.sender, order=self.order
        )

    @patch("order.signals.order_status_changed.send")
    def test_handle_order_post_save_status_unchanged(self, mock_signal):
        self.order.status = OrderStatus.PENDING
        self.order._previous_status = OrderStatus.PENDING

        handle_order_post_save(
            sender=self.sender, instance=self.order, created=False
        )

        mock_signal.assert_not_called()

    @patch("order.signals.order_status_changed.send")
    def test_handle_order_post_save_status_changed(self, mock_signal):
        self.order.status = OrderStatus.PROCESSING
        self.order._previous_status = OrderStatus.PENDING

        handle_order_post_save(
            sender=self.sender, instance=self.order, created=False
        )

        mock_signal.assert_called_once_with(
            sender=self.sender,
            order=self.order,
            old_status=OrderStatus.PENDING,
            new_status=OrderStatus.PROCESSING,
        )

    @patch("order.signals.order_status_changed.send")
    def test_handle_order_post_save_no_previous_status(self, mock_signal):
        delattr(self.order, "_previous_status")

        handle_order_post_save(
            sender=self.sender, instance=self.order, created=False
        )

        mock_signal.assert_not_called()



from decimal import Decimal
from unittest import TestCase, mock

from django.conf import settings
from djmoney.money import Money

from order.shipping import (
    FedExCarrier,
    ShippingMethodType,
    ShippingOption,
    ShippingService,
    UPSCarrier,
    get_shipping_carrier,
)


class ShippingMethodTypeTestCase(TestCase):
    def test_shipping_method_type_enum(self):
        self.assertEqual(ShippingMethodType.STANDARD.value, "STANDARD")
        self.assertEqual(ShippingMethodType.EXPRESS.value, "EXPRESS")
        self.assertEqual(ShippingMethodType.NEXT_DAY.value, "NEXT_DAY")
        self.assertEqual(ShippingMethodType.ECONOMY.value, "ECONOMY")
        self.assertEqual(
            ShippingMethodType.INTERNATIONAL.value, "INTERNATIONAL"
        )
        self.assertEqual(ShippingMethodType.PICKUP.value, "PICKUP")


class ShippingOptionTestCase(TestCase):
    def test_shipping_option_creation(self):
        option = ShippingOption(
            id="test_id",
            name="Test Option",
            method_type=ShippingMethodType.STANDARD,
            price=Money(
                amount=Decimal("10.00"), currency=settings.DEFAULT_CURRENCY
            ),
            estimated_delivery_min=3,
            estimated_delivery_max=5,
            carrier="Test Carrier",
            carrier_service_code="TEST_CODE",
            description="Test description",
        )

        self.assertEqual(option.id, "test_id")
        self.assertEqual(option.name, "Test Option")
        self.assertEqual(option.method_type, ShippingMethodType.STANDARD)
        self.assertEqual(option.price.amount, Decimal("10.00"))
        self.assertEqual(str(option.price.currency), settings.DEFAULT_CURRENCY)
        self.assertEqual(option.estimated_delivery_min, 3)
        self.assertEqual(option.estimated_delivery_max, 5)
        self.assertEqual(option.carrier, "Test Carrier")
        self.assertEqual(option.carrier_service_code, "TEST_CODE")
        self.assertEqual(option.description, "Test description")

    def test_estimated_delivery_date(self):
        option = ShippingOption(
            id="test_id",
            name="Test Option",
            method_type=ShippingMethodType.STANDARD,
            price=Money(
                amount=Decimal("10.00"), currency=settings.DEFAULT_CURRENCY
            ),
            estimated_delivery_min=3,
            estimated_delivery_max=5,
            carrier="Test Carrier",
            carrier_service_code="TEST_CODE",
        )

        min_date = option.estimated_delivery_date_min
        max_date = option.estimated_delivery_date_max

        self.assertEqual((max_date - min_date).days, 2)


@mock.patch("order.shipping.settings")
class FedExCarrierTestCase(TestCase):
    def setUp(self):
        self.from_country = mock.MagicMock()
        self.from_country.alpha_2 = "US"
        self.to_country = mock.MagicMock()
        self.to_country.alpha_2 = "US"
        self.to_region = mock.MagicMock()

        self.weight = Decimal("2.5")
        self.dimensions = {
            "length": Decimal("10"),
            "width": Decimal("5"),
            "height": Decimal("3"),
        }

    def test_init(self, mock_settings):
        mock_settings.FEDEX_API_KEY = "test_api_key"
        mock_settings.FEDEX_ACCOUNT_NUMBER = "test_account_number"

        carrier = FedExCarrier()

        self.assertEqual(carrier.api_key, "test_api_key")
        self.assertEqual(carrier.account_number, "test_account_number")

    @mock.patch("order.shipping.logger")
    def test_get_shipping_options_domestic(self, mock_logger, mock_settings):
        mock_settings.DEFAULT_CURRENCY = "EUR"

        carrier = FedExCarrier()

        options = carrier.get_shipping_options(
            weight=self.weight,
            dimensions=self.dimensions,
            from_country=self.from_country,
            to_country=self.to_country,
            to_region=self.to_region,
            to_postal_code="12345",
        )

        self.assertEqual(len(options), 3)
        self.assertEqual(options[0].carrier, "FedEx")
        self.assertEqual(options[0].method_type, ShippingMethodType.STANDARD)
        self.assertEqual(options[1].method_type, ShippingMethodType.EXPRESS)
        self.assertEqual(options[2].method_type, ShippingMethodType.NEXT_DAY)

        mock_logger.info.assert_called_once()

    @mock.patch("order.shipping.logger")
    def test_get_shipping_options_international(
        self, mock_logger, mock_settings
    ):
        mock_settings.DEFAULT_CURRENCY = "EUR"

        carrier = FedExCarrier()
        international_country = mock.MagicMock()
        international_country.alpha_2 = "GB"

        options = carrier.get_shipping_options(
            weight=self.weight,
            dimensions=self.dimensions,
            from_country=self.from_country,
            to_country=international_country,
            to_region=self.to_region,
            to_postal_code="SW1A 1AA",
        )

        self.assertEqual(len(options), 1)
        self.assertEqual(options[0].carrier, "FedEx")
        self.assertEqual(
            options[0].method_type, ShippingMethodType.INTERNATIONAL
        )

        mock_logger.info.assert_called_once()

    @mock.patch("order.shipping.FedExCarrier.create_shipment")
    @mock.patch("order.shipping.logger")
    def test_create_shipment(
        self, mock_logger, mock_create_shipment, mock_settings
    ):
        mock_create_shipment.return_value = (
            True,
            {
                "tracking_number": "FX123456789",
                "carrier": "FedEx",
                "shipping_method": "GROUND",
            },
        )

        carrier = FedExCarrier()
        order_id = "test_order_id"
        shipping_option_id = "fedex_ground"
        address_data = {
            "name": "John Doe",
            "address1": "123 Main St",
            "city": "Anytown",
            "state": "CA",
            "postal_code": "12345",
            "country": "US",
        }

        success, shipment_data = carrier.create_shipment(
            order_id=order_id,
            shipping_option_id=shipping_option_id,
            address_data=address_data,
        )

        self.assertTrue(success)
        self.assertEqual(shipment_data["tracking_number"][:2], "FX")
        self.assertEqual(shipment_data["carrier"], "FedEx")
        self.assertEqual(shipment_data["shipping_method"], "GROUND")

        mock_create_shipment.assert_called_once()

    @mock.patch("order.shipping.logger")
    def test_get_tracking_info(self, mock_logger, mock_settings):
        carrier = FedExCarrier()
        tracking_number = "FEDEX123456789"

        tracking_info = carrier.get_tracking_info(tracking_number)

        self.assertEqual(tracking_info["carrier"], "FedEx")
        self.assertEqual(tracking_info["tracking_number"], tracking_number)
        self.assertIn("status", tracking_info)
        self.assertIn("estimated_delivery", tracking_info)

        mock_logger.info.assert_called_once()


class ShippingServiceTestCase(TestCase):
    def setUp(self):
        self.from_country = mock.MagicMock()
        self.from_country.alpha_2 = "US"
        self.to_country = mock.MagicMock()
        self.to_country.alpha_2 = "US"
        self.to_region = mock.MagicMock()

        self.weight = Decimal("2.5")
        self.dimensions = {
            "length": Decimal("10"),
            "width": Decimal("5"),
            "height": Decimal("3"),
        }

    @mock.patch("order.shipping.get_shipping_carrier")
    def test_get_available_shipping_options(self, mock_get_carrier):
        fedex_carrier = mock.MagicMock()
        fedex_options = [
            ShippingOption(
                id="fedex_ground",
                name="FedEx Ground",
                method_type=ShippingMethodType.STANDARD,
                price=Money(amount="12.99", currency=settings.DEFAULT_CURRENCY),
                estimated_delivery_min=3,
                estimated_delivery_max=5,
                carrier="FedEx",
                carrier_service_code="GROUND",
            )
        ]
        fedex_carrier.get_shipping_options.return_value = fedex_options

        ups_carrier = mock.MagicMock()
        ups_options = [
            ShippingOption(
                id="ups_ground",
                name="UPS Ground",
                method_type=ShippingMethodType.STANDARD,
                price=Money(amount="11.99", currency=settings.DEFAULT_CURRENCY),
                estimated_delivery_min=2,
                estimated_delivery_max=4,
                carrier="UPS",
                carrier_service_code="GROUND",
            )
        ]
        ups_carrier.get_shipping_options.return_value = ups_options

        mock_get_carrier.side_effect = (
            lambda carrier: fedex_carrier if carrier == "fedex" else ups_carrier
        )

        options = ShippingService.get_available_shipping_options(
            order_weight=self.weight,
            order_dimensions=self.dimensions,
            from_country=self.from_country,
            to_country=self.to_country,
            to_region=self.to_region,
            to_postal_code="12345",
        )

        self.assertEqual(len(options), 2)

        self.assertEqual(options[0].name, "UPS Ground")
        self.assertEqual(options[1].name, "FedEx Ground")

    @mock.patch("order.shipping.get_shipping_carrier")
    def test_create_shipment(self, mock_get_carrier):
        carrier = mock.MagicMock()
        carrier.create_shipment.return_value = (
            True,
            {"tracking_number": "TEST123"},
        )
        mock_get_carrier.return_value = carrier

        success, shipment_data = ShippingService.create_shipment(
            order_id="test_order_id",
            carrier="fedex",
            shipping_option_id="fedex_ground",
            address_data={},
        )

        self.assertTrue(success)
        self.assertEqual(shipment_data["tracking_number"], "TEST123")
        carrier.create_shipment.assert_called_once()

    @mock.patch("order.shipping.get_shipping_carrier")
    def test_get_tracking_info(self, mock_get_carrier):
        carrier = mock.MagicMock()
        carrier.get_tracking_info.return_value = {
            "tracking_number": "TEST123",
            "status": "In Transit",
        }
        mock_get_carrier.return_value = carrier

        tracking_info = ShippingService.get_tracking_info(
            tracking_number="TEST123", carrier="fedex"
        )

        self.assertEqual(tracking_info["tracking_number"], "TEST123")
        self.assertEqual(tracking_info["status"], "In Transit")
        carrier.get_tracking_info.assert_called_once()


class GetShippingCarrierTestCase(TestCase):
    def test_get_valid_carrier(self):
        carrier = get_shipping_carrier("fedex")
        self.assertIsInstance(carrier, FedExCarrier)

        carrier = get_shipping_carrier("ups")
        self.assertIsInstance(carrier, UPSCarrier)

    def test_get_invalid_carrier(self):
        with self.assertRaises(ValueError):
            get_shipping_carrier("invalid_carrier")



from __future__ import annotations

from django.db.models.signals import post_save
from django.dispatch import receiver

from notification.models.user import NotificationUser
from notification.tasks import send_notification_task


@receiver(post_save, sender=NotificationUser)
def handle_notification_created(
    sender, instance: NotificationUser, created, **kwargs
):
    if created:
        translations_queryset = instance.notification.translations.all().values(
            "language_code", "title", "message"
        )

        translations = {
            translation["language_code"]: {
                "title": translation["title"],
                "message": translation["message"],
            }
            for translation in translations_queryset
        }

        data = {
            "user_id": instance.user.id,
            "seen": instance.seen,
            "link": instance.notification.link,
            "kind": instance.notification.kind,
            "translations": translations,
        }

        send_notification_task.delay_on_commit(data)



from django.contrib.postgres.indexes import BTreeIndex
from django.db import models
from django.utils.translation import gettext_lazy as _
from django_stubs_ext.db.models import TypedModelMeta
from mptt.fields import TreeForeignKey
from mptt.models import MPTTModel
from parler.models import TranslatableModel, TranslatedFields

from blog.managers.comment import BlogCommentManager
from core.models import TimeStampMixinModel, UUIDModel

CONTENT_PREVIEW_LENGTH = 50


class BlogComment(TranslatableModel, TimeStampMixinModel, UUIDModel, MPTTModel):
    id = models.BigAutoField(primary_key=True)
    approved = models.BooleanField(_("Approved"), default=False)
    likes = models.ManyToManyField(
        "user.UserAccount", related_name="liked_blog_comments", blank=True
    )
    user = models.ForeignKey(
        "user.UserAccount",
        related_name="blog_comments",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )
    post = models.ForeignKey(
        "blog.BlogPost",
        related_name="comments",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )
    parent = TreeForeignKey(
        "self",
        blank=True,
        null=True,
        related_name="children",
        on_delete=models.CASCADE,
    )
    translations = TranslatedFields(
        content=models.TextField(
            _("Content"),
            max_length=1000,
            blank=True,
            null=True,
        )
    )

    objects: BlogCommentManager = BlogCommentManager()

    class Meta(TypedModelMeta):
        verbose_name = _("Blog Comment")
        verbose_name_plural = _("Blog Comments")
        ordering = ["-created_at"]
        indexes = [
            *TimeStampMixinModel.Meta.indexes,
            BTreeIndex(fields=["approved"]),
            BTreeIndex(fields=["post"]),
            BTreeIndex(fields=["user"]),
            BTreeIndex(fields=["parent"]),
        ]

    class MPTTMeta:
        order_insertion_by = ["-created_at"]

    def __str__(self):
        translation_content = (
            self.safe_translation_getter("content", any_language=True)
            or "No content"
        )
        content = (
            f"{translation_content[:CONTENT_PREVIEW_LENGTH]}..."
            if len(translation_content) > CONTENT_PREVIEW_LENGTH
            else translation_content
        )
        commenter = self.user.full_name if self.user else "Anonymous"
        return f"Comment by {commenter}: {content}"

    @property
    def likes_count(self) -> int:
        return self.likes.count()

    @property
    def replies_count(self) -> int:
        return self.get_children().filter(approved=True).count()

    @property
    def approved_descendants_count(self) -> int:
        return self.get_descendants().filter(approved=True).count()



# Generated by Django 4.2.9 on 2024-02-10 14:39
import uuid
from decimal import Decimal

import django.db.models.deletion
import djmoney.models.fields
import phonenumber_field.modelfields
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Order",
            fields=[
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, verbose_name="Created At"),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, verbose_name="Updated At"),
                ),
                (
                    "uuid",
                    models.UUIDField(default=uuid.uuid4, editable=False, unique=True),
                ),
                ("deleted_at", models.DateTimeField(blank=True, null=True)),
                ("is_deleted", models.BooleanField(default=False)),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "floor",
                    models.CharField(
                        blank=True,
                        choices=[
                            (0, "Basement"),
                            (1, "Ground Floor"),
                            (2, "First Floor"),
                            (3, "Second Floor"),
                            (4, "Third Floor"),
                            (5, "Fourth Floor"),
                            (6, "Fifth Floor"),
                            (7, "Sixth Floor Plus"),
                        ],
                        default=None,
                        max_length=50,
                        null=True,
                    ),
                ),
                (
                    "location_type",
                    models.CharField(
                        blank=True,
                        choices=[(0, "Home"), (1, "Office"), (2, "Other")],
                        default=None,
                        max_length=100,
                        null=True,
                    ),
                ),
                (
                    "email",
                    models.EmailField(max_length=255, verbose_name="Email"),
                ),
                (
                    "first_name",
                    models.CharField(max_length=255, verbose_name="First Name"),
                ),
                (
                    "last_name",
                    models.CharField(max_length=255, verbose_name="Last Name"),
                ),
                (
                    "street",
                    models.CharField(max_length=255, verbose_name="Street"),
                ),
                (
                    "street_number",
                    models.CharField(max_length=255, verbose_name="Street Number"),
                ),
                (
                    "city",
                    models.CharField(max_length=255, verbose_name="City"),
                ),
                (
                    "zipcode",
                    models.CharField(max_length=255, verbose_name="Zipcode"),
                ),
                (
                    "place",
                    models.CharField(
                        blank=True,
                        max_length=255,
                        null=True,
                        verbose_name="Place",
                    ),
                ),
                (
                    "phone",
                    phonenumber_field.modelfields.PhoneNumberField(
                        max_length=128,
                        region=None,
                        verbose_name="Phone Number",
                    ),
                ),
                (
                    "mobile_phone",
                    phonenumber_field.modelfields.PhoneNumberField(
                        blank=True,
                        default=None,
                        max_length=128,
                        null=True,
                        region=None,
                        verbose_name="Mobile Phone Number",
                    ),
                ),
                (
                    "customer_notes",
                    models.TextField(
                        blank=True, null=True, verbose_name="Customer Notes"
                    ),
                ),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("SENT", "Sent"),
                            ("PAID_AND_SENT", "Paid and Sent"),
                            ("CANCELED", "Canceled"),
                            ("PENDING", "Pending"),
                        ],
                        default="PENDING",
                        max_length=20,
                        verbose_name="Status",
                    ),
                ),
                (
                    "shipping_price_currency",
                    djmoney.models.fields.CurrencyField(
                        choices=[("EUR", "EUR €"), ("USD", "USD $")],
                        default="EUR",
                        editable=False,
                        max_length=3,
                    ),
                ),
                (
                    "shipping_price",
                    djmoney.models.fields.MoneyField(
                        decimal_places=2,
                        default=Decimal("0"),
                        max_digits=11,
                        verbose_name="Shipping Price",
                    ),
                ),
                (
                    "document_type",
                    models.CharField(
                        choices=[
                            ("RECEIPT", "Receipt"),
                            ("INVOICE", "Invoice"),
                        ],
                        default="RECEIPT",
                        max_length=100,
                        verbose_name="Document Type",
                    ),
                ),
                (
                    "paid_amount_currency",
                    djmoney.models.fields.CurrencyField(
                        choices=[("EUR", "EUR €"), ("USD", "USD $")],
                        default="EUR",
                        editable=False,
                        max_length=3,
                        null=True,
                    ),
                ),
                (
                    "paid_amount",
                    djmoney.models.fields.MoneyField(
                        decimal_places=2,
                        default=Decimal("0"),
                        max_digits=11,
                        null=True,
                        verbose_name="Paid Amount",
                    ),
                ),
            ],
            options={
                "verbose_name": "Order",
                "verbose_name_plural": "Orders",
                "ordering": ["-created_at"],
            },
        ),
        migrations.CreateModel(
            name="OrderItem",
            fields=[
                (
                    "sort_order",
                    models.IntegerField(
                        db_index=True,
                        editable=False,
                        null=True,
                        verbose_name="Sort Order",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, verbose_name="Created At"),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, verbose_name="Updated At"),
                ),
                (
                    "uuid",
                    models.UUIDField(default=uuid.uuid4, editable=False, unique=True),
                ),
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                (
                    "price_currency",
                    djmoney.models.fields.CurrencyField(
                        choices=[("EUR", "EUR €"), ("USD", "USD $")],
                        default="EUR",
                        editable=False,
                        max_length=3,
                    ),
                ),
                (
                    "price",
                    djmoney.models.fields.MoneyField(
                        decimal_places=2, max_digits=11, verbose_name="Price"
                    ),
                ),
                (
                    "quantity",
                    models.IntegerField(default=1, verbose_name="Quantity"),
                ),
                (
                    "order",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="order_item_order",
                        to="order.order",
                    ),
                ),
            ],
            options={
                "verbose_name": "Order Item",
                "verbose_name_plural": "Order Items",
                "ordering": ["sort_order"],
            },
        ),
    ]



from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import extend_schema_field
from parler_rest.serializers import TranslatableModelSerializer
from rest_framework import serializers

from core.api.schema import generate_schema_multi_lang
from core.utils.serializers import TranslatedFieldExtended
from user.models.subscription import SubscriptionTopic, UserSubscription


@extend_schema_field(generate_schema_multi_lang(SubscriptionTopic))
class TranslatedFieldsFieldExtend(TranslatedFieldExtended):
    pass


class SubscriptionTopicSerializer(
    TranslatableModelSerializer, serializers.ModelSerializer[SubscriptionTopic]
):
    translations = TranslatedFieldsFieldExtend(shared_model=SubscriptionTopic)
    subscriber_count = serializers.SerializerMethodField()

    class Meta:
        model = SubscriptionTopic
        fields = [
            "translations",
            "id",
            "uuid",
            "slug",
            "category",
            "is_active",
            "is_default",
            "requires_confirmation",
            "subscriber_count",
        ]
        read_only_fields = [
            "id",
            "uuid",
            "subscriber_count",
        ]

    def get_subscriber_count(self, obj) -> int:
        if not hasattr(obj, "_subscriber_count"):
            obj._subscriber_count = obj.subscribers.filter(
                status=UserSubscription.SubscriptionStatus.ACTIVE
            ).count()
        return obj._subscriber_count


class SubscriptionTopicDetailSerializer(SubscriptionTopicSerializer):
    class Meta(SubscriptionTopicSerializer.Meta):
        fields = [
            *SubscriptionTopicSerializer.Meta.fields,
            "created_at",
            "updated_at",
        ]
        read_only_fields = [
            *SubscriptionTopicSerializer.Meta.read_only_fields,
            "created_at",
            "updated_at",
        ]


class SubscriptionTopicWriteSerializer(
    TranslatableModelSerializer, serializers.ModelSerializer[SubscriptionTopic]
):
    translations = TranslatedFieldsFieldExtend(shared_model=SubscriptionTopic)

    class Meta:
        model = SubscriptionTopic
        fields = [
            "translations",
            "slug",
            "category",
            "is_active",
            "is_default",
            "requires_confirmation",
        ]

    def validate_slug(self, value: str) -> str:
        if self.instance and self.instance.slug == value:
            return value

        if SubscriptionTopic.objects.filter(slug=value).exists():
            raise serializers.ValidationError(
                _("A subscription topic with this slug already exists.")
            )
        return value


class UserSubscriptionSerializer(serializers.ModelSerializer[UserSubscription]):
    topic_details = SubscriptionTopicSerializer(source="topic", read_only=True)
    topic = serializers.PrimaryKeyRelatedField(
        queryset=SubscriptionTopic.objects.filter(is_active=True),
    )

    class Meta:
        model = UserSubscription
        fields = [
            "id",
            "user",
            "topic",
            "topic_details",
            "status",
            "subscribed_at",
            "unsubscribed_at",
            "metadata",
            "created_at",
            "updated_at",
        ]
        read_only_fields = [
            "id",
            "user",
            "topic_details",
            "subscribed_at",
            "unsubscribed_at",
            "created_at",
            "updated_at",
        ]

    def validate_topic(self, value: SubscriptionTopic) -> SubscriptionTopic:
        try:
            topic = SubscriptionTopic.objects.get(id=value.id)
            if not topic.is_active:
                raise serializers.ValidationError(
                    _("This topic is not available for subscription.")
                )
        except SubscriptionTopic.DoesNotExist as e:
            raise serializers.ValidationError(_("Invalid topic ID.")) from e

        if self.instance is None:
            request = self.context.get("request")
            if request and hasattr(request, "user"):
                existing = UserSubscription.objects.filter(
                    user=request.user, topic=value
                ).first()
                if existing:
                    if (
                        existing.status
                        == UserSubscription.SubscriptionStatus.ACTIVE
                    ):
                        raise serializers.ValidationError(
                            _("You are already subscribed to this topic.")
                        )
                    elif (
                        existing.status
                        == UserSubscription.SubscriptionStatus.PENDING
                    ):
                        raise serializers.ValidationError(
                            _(
                                "Your subscription to this topic is pending confirmation."
                            )
                        )

        return value


class UserSubscriptionDetailSerializer(UserSubscriptionSerializer):
    class Meta(UserSubscriptionSerializer.Meta):
        fields = [
            *UserSubscriptionSerializer.Meta.fields,
            "confirmation_token",
        ]
        read_only_fields = [
            *UserSubscriptionSerializer.Meta.read_only_fields,
            "confirmation_token",
        ]


class UserSubscriptionWriteSerializer(
    serializers.ModelSerializer[UserSubscription]
):
    topic = serializers.PrimaryKeyRelatedField(
        queryset=SubscriptionTopic.objects.filter(is_active=True),
    )

    class Meta:
        model = UserSubscription
        fields = [
            "topic",
            "status",
            "metadata",
        ]

    def validate_topic(self, value: SubscriptionTopic) -> SubscriptionTopic:
        try:
            topic = SubscriptionTopic.objects.get(id=value.id)
            if not topic.is_active:
                raise serializers.ValidationError(
                    _("This topic is not available for subscription.")
                )
        except SubscriptionTopic.DoesNotExist as e:
            raise serializers.ValidationError(_("Invalid topic ID.")) from e

        if self.instance is None:
            request = self.context.get("request")
            if request and hasattr(request, "user"):
                existing = UserSubscription.objects.filter(
                    user=request.user, topic=value
                ).first()
                if existing:
                    if (
                        existing.status
                        == UserSubscription.SubscriptionStatus.ACTIVE
                    ):
                        raise serializers.ValidationError(
                            _("You are already subscribed to this topic.")
                        )
                    elif (
                        existing.status
                        == UserSubscription.SubscriptionStatus.PENDING
                    ):
                        raise serializers.ValidationError(
                            _(
                                "Your subscription to this topic is pending confirmation."
                            )
                        )

        return value


class BulkSubscriptionSerializer(serializers.Serializer):
    topic_ids = serializers.ListField(
        child=serializers.IntegerField(),
        allow_empty=False,
        help_text=_("List of topic IDs to subscribe/unsubscribe"),
    )
    action = serializers.ChoiceField(
        choices=["subscribe", "unsubscribe"],
        help_text=_("Action to perform on the topics"),
    )

    def validate_topic_ids(self, value: list[int]) -> list[int]:
        existing_ids = set(
            SubscriptionTopic.objects.filter(
                id__in=value, is_active=True
            ).values_list("id", flat=True)
        )

        invalid_ids = set(value) - existing_ids
        if invalid_ids:
            raise serializers.ValidationError(
                _("Invalid or inactive topic IDs: {}").format(list(invalid_ids))
            )

        return value


class UserSubscriptionStatusSerializer(serializers.Serializer):
    subscribed = serializers.ListField(
        child=SubscriptionTopicSerializer(), read_only=True, required=False
    )
    available = serializers.ListField(
        child=SubscriptionTopicSerializer(), read_only=True, required=False
    )

    def create(self, validated_data):
        pass

    def update(self, instance, validated_data):
        pass



# Generated by Django 4.2.9 on 2024-02-10 14:39
import uuid

import django.db.models.deletion
import parler.fields
import parler.models
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("country", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Region",
            fields=[
                (
                    "sort_order",
                    models.IntegerField(
                        db_index=True,
                        editable=False,
                        null=True,
                        verbose_name="Sort Order",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, verbose_name="Created At"),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, verbose_name="Updated At"),
                ),
                (
                    "uuid",
                    models.UUIDField(default=uuid.uuid4, editable=False, unique=True),
                ),
                (
                    "alpha",
                    models.CharField(
                        max_length=10,
                        primary_key=True,
                        serialize=False,
                        unique=True,
                        verbose_name="Region Code",
                    ),
                ),
                (
                    "country",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="regions",
                        to="country.country",
                    ),
                ),
            ],
            options={
                "verbose_name": "Region",
                "verbose_name_plural": "Regions",
                "ordering": ["sort_order"],
            },
            bases=(parler.models.TranslatableModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name="RegionTranslation",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "language_code",
                    models.CharField(
                        db_index=True, max_length=15, verbose_name="Language"
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        blank=True,
                        max_length=100,
                        null=True,
                        verbose_name="Name",
                    ),
                ),
                (
                    "master",
                    parler.fields.TranslationsForeignKey(
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="translations",
                        to="region.region",
                    ),
                ),
            ],
            options={
                "verbose_name": "Region Translation",
                "db_table": "region_region_translation",
                "db_tablespace": "",
                "managed": True,
                "default_permissions": (),
                "unique_together": {("language_code", "master")},
            },
            bases=(parler.models.TranslatedFieldsModelMixin, models.Model),
        ),
    ]



import time

from django.core.management.base import BaseCommand
from django.db.utils import OperationalError


class Command(BaseCommand):
    def handle(self, *args, **options):
        self.stdout.write("Waiting for database...")
        db_up = False
        while db_up is False:
            try:
                self.check(databases=["default"])
                db_up = True
            except OperationalError:
                self.stdout.write("Database unavailable, waiting 1 second...")
                time.sleep(1)

        self.stdout.write(self.style.SUCCESS("Database available!"))



# Generated by Django 5.1 on 2024-08-24 22:51
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    dependencies = [
        ("product", "0017_product_changed_by_historicalproduct"),
    ]

    operations = [
        migrations.AlterField(
            model_name="productcategory",
            name="sort_order",
            field=models.IntegerField(
                editable=False, null=True, verbose_name="Sort Order"
            ),
        ),
        migrations.AlterField(
            model_name="productimage",
            name="sort_order",
            field=models.IntegerField(
                editable=False, null=True, verbose_name="Sort Order"
            ),
        ),
    ]



# Generated by Django 4.2.9 on 2024-02-10 14:39
import django.db.models.deletion
from django.conf import settings
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("cart", "0002_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddField(
            model_name="cart",
            name="user",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="user_cart",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddConstraint(
            model_name="cartitem",
            constraint=models.UniqueConstraint(
                fields=("cart", "product"), name="unique_cart_item"
            ),
        ),
        migrations.AddConstraint(
            model_name="cart",
            constraint=models.UniqueConstraint(
                fields=("user",), name="unique_user_cart"
            ),
        ),
    ]



# Generated by Django 5.2 on 2025-05-08 21:53

import django.contrib.postgres.indexes
import django.db.models.deletion
import uuid
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('country', '0005_alter_country_alpha_2_alter_country_alpha_3_and_more'),
        ('order', '0009_alter_order_customer_notes_alter_order_floor_and_more'),
        ('pay_way', '0006_alter_payway_sort_order'),
        ('product', '0021_alter_historicalproduct_seo_description_and_more'),
        ('region', '0005_alter_region_sort_order'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='OrderHistory',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created At')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Updated At')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, unique=True)),
                ('change_type', models.CharField(choices=[('STATUS', 'Status Change'), ('PAYMENT', 'Payment Update'), ('SHIPPING', 'Shipping Update'), ('CUSTOMER', 'Customer Info Update'), ('ITEMS', 'Items Update'), ('ADDRESS', 'Address Update'), ('NOTE', 'Note Added'), ('REFUND', 'Refund Processed'), ('OTHER', 'Other Change')], max_length=20, verbose_name='Change Type')),
                ('previous_value', models.JSONField(blank=True, help_text='Previous value of the changed field(s).', null=True, verbose_name='Previous Value')),
                ('new_value', models.JSONField(blank=True, help_text='New value of the changed field(s).', null=True, verbose_name='New Value')),
                ('description', models.TextField(blank=True, default='', help_text='Description of what changed.', verbose_name='Description')),
                ('ip_address', models.GenericIPAddressField(blank=True, help_text='IP address from which the change was made.', null=True, verbose_name='IP Address')),
                ('user_agent', models.TextField(blank=True, help_text='User agent of the browser/client that made the change.', verbose_name='User Agent')),
            ],
            options={
                'verbose_name': 'Order History',
                'verbose_name_plural': 'Order History',
                'ordering': ['-created_at'],
            },
        ),
        migrations.CreateModel(
            name='OrderItemHistory',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created At')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Updated At')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, unique=True)),
                ('change_type', models.CharField(choices=[('QUANTITY', 'Quantity Change'), ('PRICE', 'Price Update'), ('REFUND', 'Item Refund'), ('OTHER', 'Other Change')], max_length=20, verbose_name='Change Type')),
                ('previous_value', models.JSONField(blank=True, help_text='Previous value of the changed field(s).', null=True, verbose_name='Previous Value')),
                ('new_value', models.JSONField(blank=True, help_text='New value of the changed field(s).', null=True, verbose_name='New Value')),
                ('description', models.TextField(blank=True, default='', help_text='Description of what changed.', verbose_name='Description')),
            ],
            options={
                'verbose_name': 'Order Item History',
                'verbose_name_plural': 'Order Item History',
                'ordering': ['-created_at'],
            },
        ),
        migrations.AddField(
            model_name='order',
            name='payment_id',
            field=models.CharField(blank=True, default='', max_length=255, verbose_name='Payment ID'),
        ),
        migrations.AddField(
            model_name='order',
            name='payment_method',
            field=models.CharField(blank=True, default='', max_length=50, verbose_name='Payment Method'),
        ),
        migrations.AddField(
            model_name='order',
            name='payment_status',
            field=models.CharField(blank=True, choices=[('PENDING', 'Pending'), ('PROCESSING', 'Processing'), ('COMPLETED', 'Completed'), ('FAILED', 'Failed'), ('REFUNDED', 'Refunded'), ('PARTIALLY_REFUNDED', 'Partially Refunded'), ('CANCELED', 'Canceled')], default='PENDING', max_length=50, verbose_name='Payment Status'),
        ),
        migrations.AddField(
            model_name='order',
            name='shipping_carrier',
            field=models.CharField(blank=True, default='', max_length=255, verbose_name='Shipping Carrier'),
        ),
        migrations.AddField(
            model_name='order',
            name='status_updated_at',
            field=models.DateTimeField(blank=True, null=True, verbose_name='Status Updated At'),
        ),
        migrations.AddField(
            model_name='order',
            name='tracking_number',
            field=models.CharField(blank=True, default='', max_length=255, verbose_name='Tracking Number'),
        ),
        migrations.AddField(
            model_name='orderitem',
            name='is_refunded',
            field=models.BooleanField(default=False, verbose_name='Is Refunded'),
        ),
        migrations.AddField(
            model_name='orderitem',
            name='notes',
            field=models.TextField(blank=True, default='', verbose_name='Notes'),
        ),
        migrations.AddField(
            model_name='orderitem',
            name='original_quantity',
            field=models.IntegerField(blank=True, null=True, verbose_name='Original Quantity'),
        ),
        migrations.AddField(
            model_name='orderitem',
            name='refunded_quantity',
            field=models.IntegerField(default=0, verbose_name='Refunded Quantity'),
        ),
        migrations.AlterField(
            model_name='order',
            name='document_type',
            field=models.CharField(choices=[('RECEIPT', 'Receipt'), ('INVOICE', 'Invoice'), ('PROFORMA', 'Proforma Invoice'), ('SHIPPING_LABEL', 'Shipping Label'), ('RETURN_LABEL', 'Return Label'), ('CREDIT_NOTE', 'Credit Note')], default='RECEIPT', max_length=100, verbose_name='Document Type'),
        ),
        migrations.AlterField(
            model_name='order',
            name='status',
            field=models.CharField(choices=[('PENDING', 'Pending'), ('PROCESSING', 'Processing'), ('SHIPPED', 'Shipped'), ('DELIVERED', 'Delivered'), ('COMPLETED', 'Completed'), ('CANCELED', 'Canceled'), ('RETURNED', 'Returned'), ('REFUNDED', 'Refunded')], default='PENDING', max_length=20, verbose_name='Status'),
        ),
        migrations.AddIndex(
            model_name='order',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['status_updated_at'], name='order_order_status__c9f5b7_btree'),
        ),
        migrations.AddIndex(
            model_name='order',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['payment_status'], name='order_order_payment_c12f17_btree'),
        ),
        migrations.AddIndex(
            model_name='orderitem',
            index=models.Index(fields=['product'], name='order_order_product_ed1645_idx'),
        ),
        migrations.AddIndex(
            model_name='orderitem',
            index=models.Index(fields=['is_refunded'], name='order_order_is_refu_38a302_idx'),
        ),
        migrations.AddField(
            model_name='orderhistory',
            name='order',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='history', to='order.order'),
        ),
        migrations.AddField(
            model_name='orderhistory',
            name='user',
            field=models.ForeignKey(blank=True, help_text='User who made the change, if applicable.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='order_changes', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='orderitemhistory',
            name='order_item',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='history', to='order.orderitem'),
        ),
        migrations.AddField(
            model_name='orderitemhistory',
            name='user',
            field=models.ForeignKey(blank=True, help_text='User who made the change, if applicable.', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='order_item_changes', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddIndex(
            model_name='orderhistory',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['created_at'], name='orderhistory_created_at_idx'),
        ),
        migrations.AddIndex(
            model_name='orderhistory',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['updated_at'], name='orderhistory_updated_at_idx'),
        ),
        migrations.AddIndex(
            model_name='orderhistory',
            index=models.Index(fields=['order', 'change_type'], name='order_order_order_i_4345b0_idx'),
        ),
        migrations.AddIndex(
            model_name='orderhistory',
            index=models.Index(fields=['order', '-created_at'], name='order_order_order_i_5a04bb_idx'),
        ),
        migrations.AddIndex(
            model_name='orderitemhistory',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['created_at'], name='orderitemhistory_created_at_idx'),
        ),
        migrations.AddIndex(
            model_name='orderitemhistory',
            index=django.contrib.postgres.indexes.BTreeIndex(fields=['updated_at'], name='orderitemhistory_updated_at_idx'),
        ),
        migrations.AddIndex(
            model_name='orderitemhistory',
            index=models.Index(fields=['order_item', '-created_at'], name='order_order_order_i_b122c2_idx'),
        ),
    ]



from datetime import timedelta
from django.contrib import admin, messages
from django.contrib.contenttypes.admin import GenericTabularInline
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils import timezone
from django.utils.html import conditional_escape
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _, ngettext
from parler.admin import TranslatableAdmin
from unfold.admin import ModelAdmin
from unfold.contrib.filters.admin import DropdownFilter, RangeDateTimeFilter
from unfold.decorators import action
from unfold.enums import ActionVariant

from tag.models.tag import Tag
from tag.models.tagged_item import TaggedItem


class TagStatusFilter(DropdownFilter):
    title = _("Tag Status")
    parameter_name = "tag_status"

    def lookups(self, request, model_admin):
        return [
            ("active", _("Active Tags")),
            ("inactive", _("Inactive Tags")),
            ("used", _("Used Tags")),
            ("unused", _("Unused Tags")),
            ("popular", _("Popular (>10 uses)")),
            ("recent", _("Recently Created")),
        ]

    def queryset(self, request, queryset):
        match self.value():
            case "active":
                return queryset.filter(active=True)
            case "inactive":
                return queryset.filter(active=False)
            case "used":
                return queryset.filter(taggeditem__isnull=False).distinct()
            case "unused":
                return queryset.filter(taggeditem__isnull=True)
            case "popular":
                return queryset.annotate(
                    usage_count=models.Count("taggeditem")
                ).filter(usage_count__gt=10)
            case "recent":
                cutoff = timezone.now() - timedelta(days=7)
                return queryset.filter(created_at__gte=cutoff)
            case _:
                return queryset


class TagUsageFilter(DropdownFilter):
    title = _("Usage Range")
    parameter_name = "usage_range"

    def lookups(self, request, model_admin):
        return [
            ("single", _("Single Use (1)")),
            ("low", _("Low Usage (2-5)")),
            ("medium", _("Medium Usage (6-20)")),
            ("high", _("High Usage (21-50)")),
            ("very_high", _("Very High Usage (50+)")),
        ]

    def queryset(self, request, queryset):
        qs = queryset.annotate(usage_count=models.Count("taggeditem"))
        match self.value():
            case "single":
                return qs.filter(usage_count=1)
            case "low":
                return qs.filter(usage_count__range=(2, 5))
            case "medium":
                return qs.filter(usage_count__range=(6, 20))
            case "high":
                return qs.filter(usage_count__range=(21, 50))
            case "very_high":
                return qs.filter(usage_count__gt=50)
            case _:
                return qs


class ContentTypeFilter(DropdownFilter):
    title = _("Content Type")
    parameter_name = "content_type"

    def lookups(self, request, model_admin):
        types = (
            ContentType.objects.filter(taggeditem__isnull=False)
            .distinct()
            .values_list("id", "model")
        )
        return [(cid, model.title()) for cid, model in types]

    def queryset(self, request, queryset):
        if self.value():
            return queryset.filter(
                taggeditem__content_type_id=self.value()
            ).distinct()
        return queryset


class TagInLine(GenericTabularInline):
    model = TaggedItem
    autocomplete_fields = ["tag"]
    extra = 0
    fields = ("tag",)
    verbose_name = _("Tag")
    verbose_name_plural = _("Tags")
    classes = ["collapse"]


@admin.register(Tag)
class TagAdmin(ModelAdmin, TranslatableAdmin):
    compressed_fields = True
    warn_unsaved_form = True
    list_fullwidth = True
    list_filter_submit = True
    list_filter_sheet = True

    list_display = [
        "tag_info",
        "status_badge",
        "usage_stats",
        "content_distribution",
        "sort_display",
        "created_display",
    ]
    list_filter = [
        TagStatusFilter,
        TagUsageFilter,
        ContentTypeFilter,
        "active",
        ("created_at", RangeDateTimeFilter),
        ("updated_at", RangeDateTimeFilter),
    ]
    search_fields = ["translations__label", "id"]
    readonly_fields = (
        "id",
        "uuid",
        "sort_order",
        "created_at",
        "updated_at",
        "tag_analytics",
        "usage_analytics",
        "content_analytics",
    )
    list_per_page = 50
    ordering = ["sort_order", "-id"]
    actions = [
        "activate_tags",
        "deactivate_tags",
        "update_sort_order",
        "analyze_usage",
    ]

    fieldsets = (
        (
            _("Tag Information"),
            {"fields": ("label", "active"), "classes": ("wide",)},
        ),
        (_("Organization"), {"fields": ("sort_order",), "classes": ("wide",)}),
        (
            _("Analytics"),
            {
                "fields": (
                    "tag_analytics",
                    "usage_analytics",
                    "content_analytics",
                ),
                "classes": ("collapse",),
            },
        ),
        (
            _("System Information"),
            {
                "fields": ("id", "uuid", "created_at", "updated_at"),
                "classes": ("collapse",),
            },
        ),
    )

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .annotate(
                usage_count=models.Count("taggeditem"),
                content_types_count=models.Count(
                    "taggeditem__content_type", distinct=True
                ),
            )
            .prefetch_related("taggeditem_set__content_type")
        )

    def tag_info(self, obj):
        label = obj.safe_translation_getter("label", any_language=True) or _(
            "Unnamed Tag"
        )
        display = label[:30] + ("…" if len(label) > 30 else "")
        usage = getattr(obj, "usage_count", 0)
        usage_color = (
            "text-green-600 dark:text-green-400"
            if usage > 0
            else "text-base-500 dark:text-base-400"
        )
        usage_icon = "🏷️"
        esc_color = conditional_escape(usage_color)
        esc_icon = conditional_escape(usage_icon)
        esc_label = conditional_escape(display)
        esc_id = conditional_escape(str(obj.id))
        esc_sort = conditional_escape(str(obj.sort_order or _("No order")))

        html = (
            '<div class="text-sm">'
            f'<div class="font-medium text-base-900 dark:text-base-100 flex items-center gap-2">'
            f'<span class="{esc_color}">{esc_icon}</span>'
            f"<span>{esc_label}</span>"
            "</div>"
            f'<div class="text-base-600 dark:text-base-400">ID: {esc_id}</div>'
            f'<div class="text-xs text-base-500 dark:text-base-400">Sort: {esc_sort}</div>'
            "</div>"
        )
        return mark_safe(html)

    tag_info.short_description = _("Tag")

    def status_badge(self, obj):
        if obj.active:
            status = mark_safe(
                '<span class="inline-flex items-center px-2 py-1 text-xs font-medium '
                'bg-green-50 dark:bg-green-900 text-green-700 dark:text-green-200 rounded-full">'
                "✅ Active"
                "</span>"
            )
        else:
            status = mark_safe(
                '<span class="inline-flex items-center px-2 py-1 text-xs font-medium '
                'bg-red-50 dark:bg-red-900 text-red-700 dark:text-red-200 rounded-full">'
                "❌ Inactive"
                "</span>"
            )

        label = obj.safe_translation_getter("label", any_language=True) or ""
        length = len(label)
        if length == 0:
            label_badge = mark_safe(
                '<span class="inline-flex items-center px-2 py-1 text-xs font-medium '
                'bg-red-50 dark:bg-red-900 text-red-700 dark:text-red-200 rounded mt-1">'
                "📝 No Label"
                "</span>"
            )
        elif length < 3:
            label_badge = mark_safe(
                '<span class="inline-flex items-center px-2 py-1 text-xs font-medium '
                'bg-yellow-50 dark:bg-yellow-900 text-yellow-700 dark:text-yellow-200 rounded mt-1">'
                "📝 Short"
                "</span>"
            )
        else:
            label_badge = mark_safe(
                '<span class="inline-flex items-center px-2 py-1 text-xs font-medium '
                'bg-green-50 dark:bg-green-900 text-green-700 dark:text-green-200 rounded mt-1">'
                "📝 Good"
                "</span>"
            )

        html = (
            '<div class="space-y-1">'
            f"<div>{status}</div>"
            f"<div>{label_badge}</div>"
            "</div>"
        )
        return mark_safe(html)

    status_badge.short_description = _("Status")

    def usage_stats(self, obj):
        usage = getattr(obj, "usage_count", 0)
        if usage == 0:
            badge = mark_safe(
                '<span class="inline-flex items-center px-2 py-1 text-xs font-medium '
                'bg-base-50 dark:bg-base-800 text-base-700 dark:text-base-700 rounded">'
                "🚫 Unused"
                "</span>"
            )
        elif usage == 1:
            badge = mark_safe(
                '<span class="inline-flex items-center px-2 py-1 text-xs font-medium '
                'bg-yellow-50 dark:bg-yellow-900 text-yellow-700 dark:text-yellow-200 rounded">'
                "⚠️ Single"
                "</span>"
            )
        elif usage <= 10:
            badge = mark_safe(
                '<span class="inline-flex items-center px-2 py-1 text-xs font-medium '
                'bg-blue-50 dark:bg-blue-900 text-blue-700 dark:text-blue-200 rounded">'
                "📊 Low"
                "</span>"
            )
        elif usage <= 50:
            badge = mark_safe(
                '<span class="inline-flex items-center px-2 py-1 text-xs font-medium '
                'bg-green-50 dark:bg-green-900 text-green-700 dark:text-green-200 rounded">'
                "📈 Popular"
                "</span>"
            )
        else:
            badge = mark_safe(
                '<span class="inline-flex items-center px-2 py-1 text-xs font-medium '
                'bg-purple-50 dark:bg-purple-900 text-purple-700 dark:text-purple-200 rounded">'
                "🌟 Trending"
                "</span>"
            )

        esc_usage = conditional_escape(str(usage))
        html = (
            '<div class="text-sm">'
            f'<div class="font-medium text-base-900 dark:text-base-100">{esc_usage} uses</div>'
            f'<div class="mt-1">{badge}</div>'
            "</div>"
        )
        return mark_safe(html)

    usage_stats.short_description = _("Usage")

    def content_distribution(self, obj):
        ct_count = getattr(obj, "content_types_count", 0)
        top_map = {}
        for ti in obj.taggeditem_set.all()[:10]:
            name = ti.content_type.model
            top_map[name] = top_map.get(name, 0) + 1
        if top_map:
            top, count = max(top_map.items(), key=lambda kv: kv[1])
        else:
            top, count = _("None"), 0

        if ct_count == 0:
            div_badge = mark_safe(
                '<span class="inline-flex items-center px-2 py-1 text-xs font-medium '
                'bg-base-50 dark:bg-base-800 text-base-700 dark:text-base-700 rounded">'
                "📊 No Data"
                "</span>"
            )
        elif ct_count == 1:
            div_badge = mark_safe(
                '<span class="inline-flex items-center px-2 py-1 text-xs font-medium '
                'bg-blue-50 dark:bg-blue-900 text-blue-700 dark:text-blue-200 rounded">'
                "📌 Focused"
                "</span>"
            )
        else:
            div_badge = mark_safe(
                '<span class="inline-flex items-center px-2 py-1 text-xs font-medium '
                'bg-green-50 dark:bg-green-900 text-green-700 dark:text-green-200 rounded">'
                "🌐 Diverse"
                "</span>"
            )

        esc_ct = conditional_escape(str(ct_count))
        esc_top = conditional_escape(
            top.title() if top != _("None") else _("No usage")
        )
        esc_count = conditional_escape(str(count or ""))

        html = (
            '<div class="text-sm">'
            f'<div class="font-medium text-base-900 dark:text-base-100">{esc_ct} types</div>'
            f'<div class="text-base-600 dark:text-base-400">{esc_top}: {esc_count}</div>'
            f'<div class="mt-1">{div_badge}</div>'
            "</div>"
        )
        return mark_safe(html)

    content_distribution.short_description = _("Content Types")

    def sort_display(self, obj):
        order = obj.sort_order
        if order is None:
            cls = "bg-red-50 dark:bg-red-900 text-red-700 dark:text-red-200"
            icon, label = "❌", _("None")
        elif order == 0:
            cls = "bg-blue-50 dark:bg-blue-900 text-blue-700 dark:text-blue-200"
            icon, label = "🥇", _("First")
        else:
            cls = "bg-green-50 dark:bg-green-900 text-green-700 dark:text-green-200"
            icon, label = "📋", f"#{order}"

        esc_order = conditional_escape(
            str(order if order is not None else _("None"))
        )
        esc_cls = conditional_escape(cls)
        esc_icon = conditional_escape(icon)
        esc_label = conditional_escape(label)

        html = (
            '<div class="text-sm">'
            f'<div class="font-medium text-base-900 dark:text-base-100">{esc_order}</div>'
            f'<span class="inline-flex items-center px-2 py-1 text-xs font-medium {esc_cls} rounded gap-1">'
            f"<span>{esc_icon}</span><span>{esc_label}</span>"
            "</span>"
            "</div>"
        )
        return mark_safe(html)

    sort_display.short_description = _("Sort Order")

    def created_display(self, obj):
        d = obj.created_at.strftime("%Y-%m-%d")
        t = obj.created_at.strftime("%H:%M")
        esc_d = conditional_escape(d)
        esc_t = conditional_escape(t)
        html = (
            '<div class="text-sm">'
            f'<div class="font-medium text-base-900 dark:text-base-100">{esc_d}</div>'
            f'<div class="text-base-600 dark:text-base-400">{esc_t}</div>'
            "</div>"
        )
        return mark_safe(html)

    created_display.short_description = _("Created")

    def tag_analytics(self, obj):
        label = obj.safe_translation_getter("label", any_language=True) or ""
        chars = len(label)
        words = len(label.split())
        uses = TaggedItem.objects.filter(tag=obj).count()
        status = _("Active") if obj.active else _("Inactive")
        has_lbl = _("Yes") if label else _("No")
        sort_val = (
            obj.sort_order if obj.sort_order is not None else _("Not Set")
        )

        esc_chars = conditional_escape(str(chars))
        esc_words = conditional_escape(str(words))
        esc_uses = conditional_escape(str(uses))
        esc_status = conditional_escape(status)
        esc_has = conditional_escape(has_lbl)
        esc_sort = conditional_escape(str(sort_val))

        html = (
            '<div class="text-sm"><div class="grid grid-cols-2 gap-2">'
            f"<div><strong>Label Length:</strong></div><div>{esc_chars} chars</div>"
            f"<div><strong>Word Count:</strong></div><div>{esc_words} words</div>"
            f"<div><strong>Total Usage:</strong></div><div>{esc_uses} items</div>"
            f"<div><strong>Active Status:</strong></div><div>{esc_status}</div>"
            f"<div><strong>Has Label:</strong></div><div>{esc_has}</div>"
            f"<div><strong>Sort Position:</strong></div><div>{esc_sort}</div>"
            "</div></div>"
        )
        return mark_safe(html)

    tag_analytics.short_description = _("Tag Analytics")

    def usage_analytics(self, obj):
        items = TaggedItem.objects.filter(tag=obj)
        total = items.count()
        recent = items.filter(
            created_at__gte=timezone.now() - timedelta(days=30)
        ).count()
        types = items.values("content_type__model").distinct().count()
        rate = (
            _("High") if total > 20 else _("Medium") if total > 5 else _("Low")
        )

        esc_total = conditional_escape(str(total))
        esc_recent = conditional_escape(str(recent))
        esc_types = conditional_escape(str(types))
        esc_rate = conditional_escape(rate)

        html = (
            '<div class="text-sm"><div class="grid grid-cols-2 gap-2">'
            f"<div><strong>Total Usage:</strong></div><div>{esc_total}</div>"
            f"<div><strong>Recent Usage:</strong></div><div>{esc_recent} (30d)</div>"
            f"<div><strong>Content Types:</strong></div><div>{esc_types}</div>"
            f"<div><strong>Usage Rate:</strong></div><div>{esc_rate}</div>"
            "</div></div>"
        )
        return mark_safe(html)

    usage_analytics.short_description = _("Usage Analytics")

    def content_analytics(self, obj):
        items = TaggedItem.objects.filter(tag=obj).select_related(
            "content_type"
        )[:50]
        stats = {}
        for i in items:
            m = i.content_type.model
            stats[m] = stats.get(m, 0) + 1

        diversity = len(stats)
        top = max(stats, key=stats.get) if stats else _("None")
        top_count = stats.get(top, 0)
        spec = (
            _("Specialized")
            if diversity == 1
            else _("General")
            if diversity > 3
            else _("Mixed")
        )

        esc_div = conditional_escape(str(diversity))
        esc_top = conditional_escape(
            top.title() if top != _("None") else _("No data")
        )
        esc_count = conditional_escape(str(top_count))
        esc_spec = conditional_escape(spec)

        html = (
            '<div class="text-sm"><div class="grid grid-cols-2 gap-2">'
            f"<div><strong>Diversity Score:</strong></div><div>{esc_div} types</div>"
            f"<div><strong>Most Used Type:</strong></div><div>{esc_top}</div>"
            f"<div><strong>Type Usage:</strong></div><div>{esc_count} items</div>"
            f"<div><strong>Specialization:</strong></div><div>{esc_spec}</div>"
            "</div></div>"
        )
        return mark_safe(html)

    content_analytics.short_description = _("Content Analytics")

    @action(
        description=_("Activate selected tags"),
        variant=ActionVariant.SUCCESS,
        icon="check_circle",
    )
    def activate_tags(self, request, queryset):
        updated = queryset.update(active=True)
        self.message_user(
            request,
            ngettext(
                _("%(count)d tag was activated."),
                _("%(count)d tags were activated."),
                updated,
            )
            % {"count": updated},
            messages.SUCCESS,
        )

    @action(
        description=_("Deactivate selected tags"),
        variant=ActionVariant.WARNING,
        icon="cancel",
    )
    def deactivate_tags(self, request, queryset):
        updated = queryset.update(active=False)
        self.message_user(
            request,
            ngettext(
                _("%(count)d tag was deactivated."),
                _("%(count)d tags were deactivated."),
                updated,
            )
            % {"count": updated},
            messages.WARNING,
        )

    @action(
        description=_("Update sort order"),
        variant=ActionVariant.INFO,
        icon="sort",
    )
    def update_sort_order(self, request, queryset):
        ordered = queryset.annotate(
            usage_count=models.Count("taggeditem")
        ).order_by("-usage_count", "translations__label")
        updated = 0
        for idx, tag in enumerate(ordered):
            tag.sort_order = idx
            tag.save(update_fields=["sort_order"])
            updated += 1
        self.message_user(
            request,
            _("Updated sort order for %(count)d tags.") % {"count": updated},
            messages.SUCCESS,
        )

    @action(
        description=_("Analyze tag usage"),
        variant=ActionVariant.PRIMARY,
        icon="analytics",
    )
    def analyze_usage(self, request, queryset):
        total = queryset.count()
        active = queryset.filter(active=True).count()
        used = queryset.filter(taggeditem__isnull=False).distinct().count()
        self.message_user(
            request,
            _(
                "Analysis complete: %(total)d total tags, %(active)d active, %(used)d in use."
            )
            % {"total": total, "active": active, "used": used},
            messages.INFO,
        )


@admin.register(TaggedItem)
class TaggedItemAdmin(ModelAdmin):
    list_fullwidth = True
    list_filter_submit = True

    list_display = [
        "tagged_item_info",
        "tag_display",
        "content_object_display",
        "content_type_badge",
        "created_display",
    ]
    list_filter = [
        ("tag", admin.RelatedOnlyFieldListFilter),
        ("content_type", admin.RelatedOnlyFieldListFilter),
        ("created_at", RangeDateTimeFilter),
    ]
    search_fields = ["tag__translations__label", "object_id"]
    readonly_fields = ("uuid", "created_at", "updated_at", "content_object")
    list_select_related = ["tag", "content_type"]
    list_per_page = 100
    ordering = ["-created_at"]

    fieldsets = (
        (
            _("Tagged Item Information"),
            {
                "fields": (
                    "tag",
                    "content_type",
                    "object_id",
                    "content_object",
                ),
                "classes": ("wide",),
            },
        ),
        (
            _("System Information"),
            {
                "fields": ("uuid", "created_at", "updated_at"),
                "classes": ("collapse",),
            },
        ),
    )

    def get_queryset(self, request):
        return (
            super()
            .get_queryset(request)
            .select_related("tag", "content_type")
            .prefetch_related("tag__translations")
        )

    def tagged_item_info(self, obj):
        esc_id = conditional_escape(str(obj.id))
        esc_obj = conditional_escape(str(obj.object_id))
        esc_uuid = conditional_escape(str(obj.uuid)[:8])
        html = (
            '<div class="text-sm">'
            f'<div class="font-medium text-base-900 dark:text-base-100">Item #{esc_id}</div>'
            f'<div class="text-base-600 dark:text-base-400">Object ID: {esc_obj}</div>'
            f'<div class="text-xs text-base-500 dark:text-base-400">UUID: {esc_uuid}</div>'
            "</div>"
        )
        return mark_safe(html)

    tagged_item_info.short_description = _("Tagged Item")

    def tag_display(self, obj):
        t = obj.tag
        label = t.safe_translation_getter("label", any_language=True) or _(
            "Unnamed Tag"
        )
        status_color = (
            "text-green-600 dark:text-green-400"
            if t.active
            else "text-red-600 dark:text-red-400"
        )
        status_icon = "✅" if t.active else "❌"
        esc_label = conditional_escape(label)
        esc_color = conditional_escape(status_color)
        esc_icon = conditional_escape(status_icon)
        esc_tid = conditional_escape(str(t.id))
        html = (
            '<div class="text-sm">'
            f'<div class="font-medium text-base-900 dark:text-base-100">{esc_label}</div>'
            f'<div class="flex items-center gap-1 {esc_color}">'
            f"<span>{esc_icon}</span><span>Tag #{esc_tid}</span>"
            "</div>"
            "</div>"
        )
        return mark_safe(html)

    tag_display.short_description = _("Tag")

    def content_object_display(self, obj):
        try:
            co = obj.content_object
            if co:
                text = str(co)
                display = text[:40] + ("…" if len(text) > 40 else "")
                esc_text = conditional_escape(display)
                esc_oid = conditional_escape(str(obj.object_id))
                html = (
                    '<div class="text-sm">'
                    f'<div class="font-medium text-base-900 dark:text-base-100">{esc_text}</div>'
                    f'<div class="text-base-600 dark:text-base-400">ID: {esc_oid}</div>'
                    "</div>"
                )
                return mark_safe(html)
            return mark_safe(
                '<div class="text-sm text-red-600 dark:text-red-400">Object not found</div>'
            )
        except Exception:
            return mark_safe(
                '<div class="text-sm text-red-600 dark:text-red-400">Error loading object</div>'
            )

    content_object_display.short_description = _("Content Object")

    def content_type_badge(self, obj):
        name = obj.content_type.model
        cfg = {
            "product": (
                "bg-green-50 dark:bg-green-900",
                "text-green-700 dark:text-green-200",
                "🏷️",
            ),
            "article": (
                "bg-blue-50 dark:bg-blue-900",
                "text-blue-700 dark:text-blue-200",
                "📄",
            ),
            "user": (
                "bg-purple-50 dark:bg-purple-900",
                "text-purple-700 dark:text-purple-200",
                "👤",
            ),
            "category": (
                "bg-orange-50 dark:bg-orange-900",
                "text-orange-700 dark:text-orange-200",
                "📁",
            ),
        }.get(
            name,
            (
                "bg-base-50 dark:bg-base-800",
                "text-base-700 dark:text-base-700",
                "📦",
            ),
        )
        bg, text, icon = cfg
        esc_bg = conditional_escape(bg)
        esc_text = conditional_escape(text)
        esc_icon = conditional_escape(icon)
        esc_name = conditional_escape(name.title())
        html = (
            '<span class="inline-flex items-center px-2 py-1 text-xs font-medium '
            f'{esc_bg} {esc_text} rounded border gap-1">'
            f"<span>{esc_icon}</span><span>{esc_name}</span>"
            "</span>"
        )
        return mark_safe(html)

    content_type_badge.short_description = _("Content Type")

    def created_display(self, obj):
        d = obj.created_at.strftime("%Y-%m-%d")
        t = obj.created_at.strftime("%H:%M")
        esc_d = conditional_escape(d)
        esc_t = conditional_escape(t)
        html = (
            '<div class="text-sm">'
            f'<div class="font-medium text-base-900 dark:text-base-100">{esc_d}</div>'
            f'<div class="text-base-600 dark:text-base-400">{esc_t}</div>'
            "</div>"
        )
        return mark_safe(html)

    created_display.short_description = _("Created")


class TaggedItemInline(GenericTabularInline):
    model = TaggedItem
    extra = 0
    fields = ("tag",)
    verbose_name = _("Tag")
    verbose_name_plural = _("Tags")
    ct_field = "content_type"
    ct_fk_field = "object_id"



from django.urls import path

from order.views.item import OrderItemViewSet
from order.views.order import OrderViewSet
from order.views.payment import OrderPaymentViewSet

urlpatterns = [
    path(
        "order",
        OrderViewSet.as_view({"get": "list", "post": "create"}),
        name="order-list",
    ),
    path(
        "order/<int:pk>",
        OrderViewSet.as_view(
            {
                "get": "retrieve",
                "put": "update",
                "patch": "partial_update",
                "delete": "destroy",
            }
        ),
        name="order-detail",
    ),
    path(
        "order/uuid/<uuid:uuid>",
        OrderViewSet.as_view({"get": "retrieve_by_uuid"}),
        name="order-retrieve-by-uuid",
    ),
    path(
        "order/<int:pk>/create_payment_intent",
        OrderViewSet.as_view({"post": "create_payment_intent"}),
        name="order-create-payment-intent",
    ),
    path(
        "order/<int:pk>/create_checkout_session",
        OrderViewSet.as_view({"post": "create_checkout_session"}),
        name="order-create-checkout-session",
    ),
    path(
        "order/<int:pk>/cancel",
        OrderViewSet.as_view({"post": "cancel"}),
        name="order-cancel",
    ),
    path(
        "order/<int:pk>/add_tracking",
        OrderViewSet.as_view({"post": "add_tracking"}),
        name="order-add-tracking",
    ),
    path(
        "order/<int:pk>/update_status",
        OrderViewSet.as_view({"post": "update_status"}),
        name="order-update-status",
    ),
    path(
        "order/my_orders",
        OrderViewSet.as_view({"get": "my_orders"}),
        name="order-my-orders",
    ),
    path(
        "order/<int:pk>/process_payment",
        OrderPaymentViewSet.as_view({"post": "process_payment"}),
        name="order-process-payment",
    ),
    path(
        "order/<int:pk>/payment_status",
        OrderPaymentViewSet.as_view({"get": "check_payment_status"}),
        name="order-payment-status",
    ),
    path(
        "order/<int:pk>/refund",
        OrderPaymentViewSet.as_view({"post": "refund_payment"}),
        name="order-refund",
    ),
    path(
        "order-items",
        OrderItemViewSet.as_view({"get": "list", "post": "create"}),
        name="order-item-list",
    ),
    path(
        "order-items/<int:pk>",
        OrderItemViewSet.as_view(
            {
                "get": "retrieve",
                "put": "update",
                "patch": "partial_update",
                "delete": "destroy",
            }
        ),
        name="order-item-detail",
    ),
    path(
        "order-items/<int:pk>/refund",
        OrderItemViewSet.as_view({"post": "refund"}),
        name="order-item-refund",
    ),
]



from unittest import mock

from allauth.socialaccount.adapter import DefaultSocialAccountAdapter
from allauth.socialaccount.models import SocialLogin
from django.contrib.auth import get_user_model
from django.test import TestCase

from user.adapter import SocialAccountAdapter

User = get_user_model()


class TestSocialAccountAdapter(TestCase):
    def setUp(self):
        self.adapter = SocialAccountAdapter()

    @mock.patch.object(DefaultSocialAccountAdapter, "pre_social_login")
    @mock.patch.object(
        SocialLogin, "is_existing", new_callable=mock.PropertyMock
    )
    def test_pre_social_login_existent(
        self, mock_is_existing, mock_pre_social_login
    ):
        request = mock.Mock()
        sociallogin = SocialLogin()
        mock_is_existing.return_value = True
        self.adapter.pre_social_login(request, sociallogin)
        mock_pre_social_login.assert_not_called()

    @mock.patch.object(DefaultSocialAccountAdapter, "pre_social_login")
    def test_pre_social_login_nonexistent(self, mock_pre_social_login):
        request = mock.Mock()

        sociallogin = mock.Mock()

        is_existing_property = mock.PropertyMock(return_value=False)
        type(sociallogin).is_existing = is_existing_property

        sociallogin.account = mock.Mock()
        sociallogin.account.extra_data = {"email": "test@example.com"}

        email = mock.Mock()
        email.verified = True
        email.email = "test@example.com"
        sociallogin.email_addresses = [email]

        mock_user = mock.Mock()

        mock_filter = mock.Mock()
        mock_filter.first.return_value = mock_user

        def side_effect(req, user):
            pass

        sociallogin.connect.side_effect = side_effect

        with mock.patch.object(
            User.objects, "filter", return_value=mock_filter
        ):
            self.adapter.pre_social_login(request, sociallogin)

            sociallogin.connect.assert_called_once_with(request, mock_user)



import os

from django.conf import settings
from django.core.files.storage import default_storage
from django.test import TestCase

from product.factories.image import ProductImageFactory
from product.factories.product import ProductFactory

languages = [
    lang["code"] for lang in settings.PARLER_LANGUAGES[settings.SITE_ID]
]
default_language = settings.PARLER_DEFAULT_LANGUAGE_CODE


class ProductImageModelTestCase(TestCase):
    def setUp(self):
        self.product = ProductFactory(num_images=0, num_reviews=0)
        self.product_image = ProductImageFactory(
            product=self.product,
            is_main=True,
        )

    def test_fields(self):
        self.assertEqual(self.product_image.product, self.product)
        self.assertTrue(self.product_image.is_main)
        self.assertTrue(default_storage.exists(self.product_image.image.path))

    def test_str_representation(self):
        product_name = self.product.safe_translation_getter(
            "name", any_language=True
        )
        main_status = "Main" if self.product_image.is_main else "Secondary"

        self.assertEqual(
            str(self.product_image),
            f"{product_name} Image ({main_status})",
        )

    def test_get_ordering_queryset(self):
        queryset = self.product_image.get_ordering_queryset()
        self.assertEqual(queryset.count(), 1)
        self.assertTrue(self.product_image in queryset)
        self.assertEqual(queryset.first(), self.product_image)

    def test_main_image_path(self):
        expected_filename = f"media/uploads/products/{os.path.basename(self.product_image.image.name)}"
        self.assertEqual(self.product_image.main_image_path, expected_filename)



from datetime import datetime

from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import extend_schema_field
from parler_rest.serializers import TranslatableModelSerializer
from rest_framework import serializers

from core.api.schema import generate_schema_multi_lang
from core.utils.serializers import TranslatedFieldExtended
from notification.models.notification import Notification


@extend_schema_field(generate_schema_multi_lang(Notification))
class TranslatedFieldsFieldExtend(TranslatedFieldExtended):
    pass


class NotificationSerializer(
    TranslatableModelSerializer, serializers.ModelSerializer[Notification]
):
    translations = TranslatedFieldsFieldExtend(shared_model=Notification)
    link = serializers.SerializerMethodField()

    @extend_schema_field(
        {
            "type": "string",
            "nullable": True,
            "maxLength": 200,
            "description": _("URL link or empty string"),
        }
    )
    def get_link(self, obj) -> str | None:
        return obj.link

    class Meta:
        model = Notification
        fields = (
            "translations",
            "id",
            "link",
            "kind",
            "expiry_date",
            "created_at",
            "updated_at",
            "uuid",
        )
        read_only_fields = (
            "id",
            "created_at",
            "updated_at",
            "uuid",
        )


class NotificationDetailSerializer(NotificationSerializer):
    class Meta(NotificationSerializer.Meta):
        fields = (*NotificationSerializer.Meta.fields,)


class NotificationWriteSerializer(
    TranslatableModelSerializer, serializers.ModelSerializer[Notification]
):
    translations = TranslatedFieldsFieldExtend(shared_model=Notification)

    def validate_expiry_date(self, value: datetime) -> datetime:
        if value and value <= timezone.now():
            raise serializers.ValidationError(
                _("Expiry date must be in the future.")
            )
        return value

    def validate_link(self, value: str) -> str:
        if value and not (
            value.startswith("http://")
            or value.startswith("https://")
            or value.startswith("/")
        ):
            raise serializers.ValidationError(
                _("Link must be a valid URL or relative path.")
            )
        return value

    class Meta:
        model = Notification
        fields = (
            "translations",
            "link",
            "kind",
            "expiry_date",
        )



from datetime import timedelta
from django.contrib.auth import get_user_model
from django.urls import reverse
from django.utils import timezone
from rest_framework.test import APITestCase

from user.factories.subscription import (
    SubscriptionTopicFactory,
    UserSubscriptionFactory,
)
from user.factories.account import UserAccountFactory
from user.models.subscription import SubscriptionTopic, UserSubscription

User = get_user_model()


class UserSubscriptionFilterTest(APITestCase):
    def setUp(self):
        UserSubscription.objects.all().delete()
        SubscriptionTopic.objects.all().delete()
        User.objects.all().delete()

        self.user = UserAccountFactory()
        self.other_user = UserAccountFactory()

        self.now = timezone.now()

        self.newsletter_topic = SubscriptionTopicFactory(
            slug="weekly-newsletter",
            category=SubscriptionTopic.TopicCategory.NEWSLETTER,
            is_active=True,
        )
        self.newsletter_topic.set_current_language("en")
        self.newsletter_topic.name = "Weekly Newsletter"
        self.newsletter_topic.description = "Get our weekly updates"
        self.newsletter_topic.save()

        self.marketing_topic = SubscriptionTopicFactory(
            slug="marketing-campaigns",
            category=SubscriptionTopic.TopicCategory.MARKETING,
            is_active=True,
        )
        self.marketing_topic.set_current_language("en")
        self.marketing_topic.name = "Marketing Updates"
        self.marketing_topic.description = "Special offers and promotions"
        self.marketing_topic.save()

        self.product_topic = SubscriptionTopicFactory(
            slug="product-updates",
            category=SubscriptionTopic.TopicCategory.PRODUCT,
            is_active=True,
        )
        self.product_topic.set_current_language("en")
        self.product_topic.name = "Product News"
        self.product_topic.description = "Latest product features"
        self.product_topic.save()

        self.active_subscription = UserSubscriptionFactory(
            user=self.user,
            topic=self.newsletter_topic,
            status=UserSubscription.SubscriptionStatus.ACTIVE,
            confirmation_token="",
            metadata={
                "source": "website",
                "preferences": {"frequency": "weekly"},
            },
        )
        self.active_subscription.created_at = self.now - timedelta(days=30)
        self.active_subscription.subscribed_at = self.now - timedelta(days=30)
        self.active_subscription.save()

        self.pending_subscription = UserSubscriptionFactory(
            user=self.user,
            topic=self.marketing_topic,
            status=UserSubscription.SubscriptionStatus.PENDING,
            confirmation_token="abc123token",
            metadata={},
        )
        self.pending_subscription.created_at = self.now - timedelta(days=15)
        self.pending_subscription.subscribed_at = self.now - timedelta(days=15)
        self.pending_subscription.save()

        self.unsubscribed_subscription = UserSubscriptionFactory(
            user=self.user,
            topic=self.product_topic,
            status=UserSubscription.SubscriptionStatus.UNSUBSCRIBED,
            confirmation_token="",
            metadata={"unsubscribe_reason": "too_frequent"},
        )
        self.unsubscribed_subscription.created_at = self.now - timedelta(days=5)
        self.unsubscribed_subscription.subscribed_at = self.now - timedelta(
            days=5
        )
        self.unsubscribed_subscription.unsubscribed_at = self.now - timedelta(
            days=2
        )
        self.unsubscribed_subscription.save()

        self.other_user_subscription = UserSubscriptionFactory(
            user=self.other_user,
            topic=self.newsletter_topic,
            status=UserSubscription.SubscriptionStatus.ACTIVE,
            confirmation_token="",
            metadata={"source": "mobile_app"},
        )
        self.other_user_subscription.created_at = self.now - timedelta(days=10)
        self.other_user_subscription.subscribed_at = self.now - timedelta(
            days=10
        )
        self.other_user_subscription.save()

        self.client.force_authenticate(user=self.user)

    def test_timestamp_filters(self):
        url = reverse("user-subscription-list")

        created_after_date = self.now - timedelta(days=20)
        response = self.client.get(
            url,
            {"created_after": created_after_date.isoformat()},
        )
        self.assertEqual(response.status_code, 200)

        result_ids = [r["id"] for r in response.data["results"]]
        self.assertNotIn(self.active_subscription.id, result_ids)
        self.assertIn(self.pending_subscription.id, result_ids)
        self.assertIn(self.unsubscribed_subscription.id, result_ids)
        self.assertEqual(len(result_ids), 2)

        created_before_date = self.now - timedelta(days=20)
        response = self.client.get(
            url,
            {"created_before": created_before_date.isoformat()},
        )
        self.assertEqual(response.status_code, 200)

        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.active_subscription.id, result_ids)

    def test_uuid_filter(self):
        url = reverse("user-subscription-list")

        response = self.client.get(
            url, {"uuid": str(self.active_subscription.uuid)}
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(len(response.data["results"]), 1)
        self.assertEqual(
            response.data["results"][0]["id"], self.active_subscription.id
        )

    def test_camel_case_filters(self):
        url = reverse("user-subscription-list")

        created_after_date = self.now - timedelta(days=20)
        response = self.client.get(
            url,
            {
                "createdAfter": created_after_date.isoformat(),
                "status": UserSubscription.SubscriptionStatus.PENDING,
            },
        )
        self.assertEqual(response.status_code, 200)

        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.pending_subscription.id, result_ids)

        response = self.client.get(
            url,
            {
                "subscribedAfter": (self.now - timedelta(days=20)).isoformat(),
                "topicCategory": SubscriptionTopic.TopicCategory.NEWSLETTER,
            },
        )
        self.assertEqual(response.status_code, 200)

        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 0)

    def test_status_filters(self):
        url = reverse("user-subscription-list")

        response = self.client.get(
            url, {"status": UserSubscription.SubscriptionStatus.ACTIVE}
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.active_subscription.id, result_ids)

        response = self.client.get(
            url, {"status": UserSubscription.SubscriptionStatus.PENDING}
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.pending_subscription.id, result_ids)

        response = self.client.get(
            url, {"status": UserSubscription.SubscriptionStatus.UNSUBSCRIBED}
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.unsubscribed_subscription.id, result_ids)

    def test_topic_filters(self):
        url = reverse("user-subscription-list")

        response = self.client.get(url, {"topic": self.newsletter_topic.id})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.active_subscription.id, result_ids)

        response = self.client.get(
            url, {"topic_category": SubscriptionTopic.TopicCategory.MARKETING}
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.pending_subscription.id, result_ids)

    def test_subscription_date_filters(self):
        url = reverse("user-subscription-list")

        subscribed_after_date = self.now - timedelta(days=20)
        response = self.client.get(
            url,
            {"subscribed_after": subscribed_after_date.isoformat()},
        )
        self.assertEqual(response.status_code, 200)

        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 2)
        self.assertIn(self.pending_subscription.id, result_ids)
        self.assertIn(self.unsubscribed_subscription.id, result_ids)

        unsubscribed_after_date = self.now - timedelta(days=3)
        response = self.client.get(
            url,
            {"unsubscribed_after": unsubscribed_after_date.isoformat()},
        )
        self.assertEqual(response.status_code, 200)

        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.unsubscribed_subscription.id, result_ids)

    def test_topic_slug_filters(self):
        url = reverse("user-subscription-list")

        response = self.client.get(url, {"topic_slug": "newsletter"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.active_subscription.id, result_ids)

        response = self.client.get(
            url, {"topic_slug_exact": "marketing-campaigns"}
        )
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.pending_subscription.id, result_ids)

    def test_topic_translation_filters(self):
        url = reverse("user-subscription-list")

        response = self.client.get(url, {"topic_name": "Newsletter"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.active_subscription.id, result_ids)

        response = self.client.get(url, {"topic_description": "offers"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.pending_subscription.id, result_ids)

    def test_custom_filters(self):
        url = reverse("user-subscription-list")

        response = self.client.get(url, {"is_confirmed": "true"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 2)
        self.assertIn(self.active_subscription.id, result_ids)
        self.assertIn(self.unsubscribed_subscription.id, result_ids)

        response = self.client.get(url, {"is_confirmed": "false"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.pending_subscription.id, result_ids)

        response = self.client.get(url, {"has_metadata": "true"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 2)
        self.assertIn(self.active_subscription.id, result_ids)
        self.assertIn(self.unsubscribed_subscription.id, result_ids)

        response = self.client.get(url, {"has_metadata": "false"})
        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.pending_subscription.id, result_ids)

    def test_complex_filter_combinations(self):
        url = reverse("user-subscription-list")

        created_after_date = self.now - timedelta(days=25)
        response = self.client.get(
            url,
            {
                "createdAfter": created_after_date.isoformat(),
                "status": UserSubscription.SubscriptionStatus.PENDING,
                "topicCategory": SubscriptionTopic.TopicCategory.MARKETING,
                "isConfirmed": "false",
                "ordering": "-createdAt",
            },
        )

        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.pending_subscription.id, result_ids)

        response = self.client.get(
            url,
            {
                "topicName": "Newsletter",
                "hasMetadata": "true",
                "status": UserSubscription.SubscriptionStatus.ACTIVE,
                "isConfirmed": "true",
            },
        )

        self.assertEqual(response.status_code, 200)
        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.active_subscription.id, result_ids)

    def test_filter_with_ordering(self):
        url = reverse("user-subscription-list")

        response = self.client.get(url, {"ordering": "-createdAt"})
        self.assertEqual(response.status_code, 200)

        results = response.data["results"]
        self.assertEqual(len(results), 3)

        subscription_ids = [r["id"] for r in results]
        self.assertEqual(subscription_ids[0], self.unsubscribed_subscription.id)
        self.assertEqual(subscription_ids[1], self.pending_subscription.id)
        self.assertEqual(subscription_ids[2], self.active_subscription.id)

        response = self.client.get(url, {"ordering": "status,-subscribedAt"})
        self.assertEqual(response.status_code, 200)

        results = response.data["results"]
        self.assertEqual(len(results), 3)
        statuses = [r["status"] for r in results]
        self.assertTrue(statuses.index("ACTIVE") < statuses.index("PENDING"))

    def test_existing_filters_still_work(self):
        url = reverse("user-subscription-list")

        response = self.client.get(
            url,
            {
                "status": UserSubscription.SubscriptionStatus.ACTIVE,
                "topic_category": SubscriptionTopic.TopicCategory.NEWSLETTER,
                "is_confirmed": "true",
            },
        )
        self.assertEqual(response.status_code, 200)

        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 1)
        self.assertIn(self.active_subscription.id, result_ids)

    def test_user_isolation(self):
        url = reverse("user-subscription-list")

        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        result_ids = [r["id"] for r in response.data["results"]]
        self.assertEqual(len(result_ids), 3)
        self.assertIn(self.active_subscription.id, result_ids)
        self.assertIn(self.pending_subscription.id, result_ids)
        self.assertIn(self.unsubscribed_subscription.id, result_ids)
        self.assertNotIn(self.other_user_subscription.id, result_ids)

        self.client.force_authenticate(user=self.other_user)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        result_ids = [r["id"] for r in response.data["results"]]
        self.assertGreaterEqual(len(result_ids), 1)
        self.assertIn(self.other_user_subscription.id, result_ids)
        self.assertNotIn(self.active_subscription.id, result_ids)
        self.assertNotIn(self.pending_subscription.id, result_ids)
        self.assertNotIn(self.unsubscribed_subscription.id, result_ids)

    def tearDown(self):
        UserSubscription.objects.all().delete()
        SubscriptionTopic.objects.all().delete()



# Generated by Django 5.0.3 on 2024-03-31 11:07
import django.contrib.postgres.indexes
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("country", "0001_initial"),
        ("region", "0001_initial"),
        ("user", "0002_alter_useraddress_options"),
    ]

    operations = [
        migrations.AddIndex(
            model_name="useraddress",
            index=django.contrib.postgres.indexes.GinIndex(
                fields=["title", "first_name", "last_name", "city"],
                name="address_search_gin",
                opclasses=[
                    "gin_trgm_ops",
                    "gin_trgm_ops",
                    "gin_trgm_ops",
                    "gin_trgm_ops",
                ],
            ),
        ),
    ]



from django.contrib.postgres.indexes import BTreeIndex
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _
from django_stubs_ext.db.models import TypedModelMeta

from core.models import TimeStampMixinModel, UUIDModel
from product.managers.favourite import FavouriteManager


class ProductFavourite(TimeStampMixinModel, UUIDModel):
    id = models.BigAutoField(primary_key=True)
    user = models.ForeignKey(
        "user.UserAccount",
        related_name="favourite_products",
        on_delete=models.CASCADE,
    )
    product = models.ForeignKey(
        "product.Product",
        related_name="favourited_by",
        on_delete=models.CASCADE,
    )

    objects: FavouriteManager = FavouriteManager()

    class Meta(TypedModelMeta):
        verbose_name = _("Product Favourite")
        verbose_name_plural = _("Product Favourites")
        ordering = ["-created_at"]
        constraints = [
            models.UniqueConstraint(
                fields=["user", "product"], name="unique_product_favourite"
            )
        ]
        indexes = [
            *TimeStampMixinModel.Meta.indexes,
            BTreeIndex(fields=["user"], name="product_favourite_user_ix"),
            BTreeIndex(fields=["product"], name="product_favourite_product_ix"),
        ]

    def __str__(self):
        product_name = self.product.safe_translation_getter(
            "name", any_language=True
        )
        return f"{self.user.email} - {product_name}"

    def save(self, *args, **kwargs):
        if (
            not self.pk
            and ProductFavourite.objects.filter(
                user=self.user, product=self.product
            ).exists()
        ):
            raise ValidationError(
                _("This product is already in the user's favorites.")
            )
        super().save(*args, **kwargs)



from __future__ import annotations
from rest_framework import serializers
from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import (
    extend_schema,
    inline_serializer,
    OpenApiParameter,
)
from rest_framework import status
from rest_framework.decorators import api_view
from rest_framework.response import Response

from notification.models.notification import Notification
from notification.models.user import NotificationUser
from notification.serializers.notification import NotificationSerializer


@extend_schema(
    operation_id="getNotificationsByIds",
    summary=_("Returns the notifications for a list of ids."),
    description=_("Returns the notifications for a list of ids."),
    tags=["Notifications"],
    parameters=[
        OpenApiParameter(
            name="seen",
            type=bool,
            location=OpenApiParameter.QUERY,
            required=False,
            description=_(
                "Filter notifications by seen status. If false, returns only unseen notifications."
            ),
        ),
    ],
    request=inline_serializer(
        name="NotificationIdsSerializer",
        fields={"ids": serializers.ListField(child=serializers.IntegerField())},
    ),
    responses=NotificationSerializer(many=True),
    methods=["POST"],
)
@api_view(["POST"])
def notifications_by_ids(request):
    seen_query = request.query_params.get("seen", None)
    notification_ids = request.data.get("ids", [])

    if not notification_ids:
        return Response(
            {"error": _("No notification ids provided.")},
            status=status.HTTP_400_BAD_REQUEST,
        )

    notifications = Notification.objects.filter(id__in=notification_ids)

    if seen_query is not None:
        user = request.user
        seen_value = seen_query.lower() == "true"

        if seen_value:
            notifications = notifications.filter(
                user__user=user, user__seen=True
            )
        else:
            seen_notification_ids = NotificationUser.objects.filter(
                user=user, notification_id__in=notification_ids, seen=True
            ).values_list("notification_id", flat=True)

            notifications = notifications.exclude(id__in=seen_notification_ids)

    if not notifications.exists():
        return Response(
            {"error": _("No notifications found for the provided ids.")},
            status=status.HTTP_404_NOT_FOUND,
        )

    serializer = NotificationSerializer(notifications, many=True)
    return Response(serializer.data)
